# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module to handle spectra data
"""

import glob
import re

import numpy
import numpy.lib.recfunctions
from numpy import rec

from .. import config
from .helper import xu_h5open

re_wspaces = re.compile(r"\s+")
re_colname = re.compile(r"^Col")

re_comment_section = re.compile(r"^%c")
re_parameter_section = re.compile(r"^%p")
re_data_section = re.compile(r"^%d")
re_end_section = re.compile(r"^!")
re_unit = re.compile(r"\[.+\]")
re_obracket = re.compile(r"\[")
re_cbracket = re.compile(r"\]")
re_underscore = re.compile(r"_")
re_column = re.compile(r"^Col")
re_col_name = re.compile(r"\d+\s+.+\s*\[")
re_col_index = re.compile(r"\d+\s+")
re_col_type = re.compile(r"\[.+\]")
re_num = re.compile(r"[0-9]")

dtype_map = {"FLOAT": "f4",
             "DOUBLE": "f8"}


class SPECTRAFileComments(dict):
    """
    Class that describes the comments in the header of a SPECTRA file.
    The different comments are accessible via the comment keys.
    """

    def __init__(self):
        pass

    def __getattr__(self, name):
        if name in self:
            return self[name]


class SPECTRAFileParameters(dict):

    def __init__(self):
        pass

    def __getattr__(self, name):
        if name in self:
            return self[name]

    def __str__(self):
        ostr = ""
        lmax_key = 0
        lmax_item = 0

        # find the length of the longest key
        for k in self:
            if len(k) > lmax_key:
                lmax_key = len(k)

            i = self[k]
            if not isinstance(i, str):
                # if the item is not a string it must be converted
                i = "%f" % i

            if len(i) > lmax_item:
                lmax_item = len(i)

        # define the format string for a single key-value pair
        kvfmt = "|%%-%is = %%-%is" % (lmax_key, lmax_item)

        cnt = 0
        ostr += (3 * (lmax_key + lmax_item + 4) + 1) * "-" + "\n"
        ostr += "|Parameters:" + (3 * (lmax_key + lmax_item)) * " " + "|\n"
        ostr += (3 * (lmax_key + lmax_item + 4) + 1) * "-" + "\n"
        for key in self:
            value = self[key]
            if not isinstance(value, str):
                value = "%f" % value

            ostr += kvfmt % (key, value)
            cnt += 1
            if cnt == 3:
                ostr += "|\n"
                cnt = 0

        if cnt != 0:
            ostr += "|\n"
        ostr += (3 * (lmax_key + lmax_item + 4) + 1) * "-" + "\n"

        return ostr


class SPECTRAFileDataColumn(object):

    def __init__(self, index, name, unit, type):
        self.index = int(index)
        self.name = name
        self.unit = unit
        self.type = type

    def __str__(self):
        ostr = "%i %s %s %s" % (self.index, self.name, self.unit, self.type)
        return ostr


class SPECTRAFileData(object):

    def __init__(self):
        self.collist = []
        self.data = None

    def append(self, col):
        self.collist.append(col)

    def __getitem__(self, key):
        try:
            return self.data[key]
        except IndexError:
            print("XU.io.specta.SPECTRAFileData: data contains no column "
                  "named: %s!" % key)

    def __str__(self):
        ostr = ""

        # determine the maximum lenght of every column string
        lmax = 0
        for c in self.collist:
            if len(c.__str__()) > lmax:
                lmax = len(c.__str__())

        lmax += 3

        # want to print in three columns
        nc = 3
        nres = len(self.collist) % nc
        nrows = (len(self.collist) - nres) / nc

        fmtstr = "| %%-%is| %%-%is| %%-%is|\n" % (lmax, lmax, lmax)

        ostr += (3 * lmax + 7) * "-" + "\n"
        ostr += "|Column names:" + (3 * lmax - 8) * " " + "|\n"
        ostr += (3 * lmax + 7) * "-" + "\n"
        # full output rows
        for i in range(nrows):
            c1 = self.collist[i * nc + 0]
            c2 = self.collist[i * nc + 1]
            c3 = self.collist[i * nc + 2]
            ostr += fmtstr % (c1.__str__(), c2.__str__(), c3.__str__())

        # residual output row
        c = ['', '', '']
        for j in range(nres):
            c[j] = self.collist[-nres + j]

        ostr += fmtstr % (c[0].__str__(), c[1].__str__(), c[2].__str__())

        ostr += (3 * lmax + 7) * "-" + "\n"
        return ostr


class SPECTRAFile(object):

    """
    Represents a SPECTRA data file. The file is read during the
    Constructor call. This class should work for data stored at
    beamlines P08 and BW2 at HASYLAB.

    Parameters
    ----------
    filename :  str
        a string with the name of the SPECTRA file

    mcatmp :    str, optional
        template for the MCA files
    mcastart, mcastop : int, optional
        start and stop index for the MCA files, if not given, the class tries
        to determine the start and stop index automatically.
    """

    def __init__(self, filename, mcatmp=None, mcastart=None, mcastop=None):
        self.filename = filename
        self.comments = SPECTRAFileComments()
        self.params = SPECTRAFileParameters()
        self.data = SPECTRAFileData()
        self.mca = None
        self.mca_channels = None

        self.Read()  # reads the .fio data file

        if mcatmp is not None:
            self.mca_file_template = mcatmp

            if mcastart is not None and mcastop is not None:
                self.mca_start_index = mcastart
                self.mca_stop_index = mcastop
            else:
                # try to determine the number of MCA spectra automatically
                spat = self.mca_file_template.replace("%i", "*")
                lst = glob.glob(spat)
                self.mca_start_index = 1
                self.mca_stop_index = 0
                if lst:
                    self.mca_stop_index = self.data.data.size  # len(l)

            if self.mca_stop_index != 0:
                self.ReadMCA()

    def Save2HDF5(self, h5file, name, group="/", mcaname="MCA"):
        """
        Saves the scan to an HDF5 file. The scan is saved to a
        seperate group of name "name". h5file is either a string
        for the file name or a HDF5 file object.
        If the mca attribute is not None mca data will be stored to an
        chunked array of with name mcaname.

        Parameters
        ----------
        h5file :    file-handle or str
            HDF5 file object or name
        name :	    str
            name of the group where to store the data

        group :	    str, optional
            root group where to store the data
        mcaname :   str, optional
            Name of the MCA in the HDF5 file

        Returns
        -------
        bool or None
            The method returns None in the case of everything went fine, True
            otherwise.
        """
        with xu_h5open(h5file, 'w') as h5:
            # create the group where to store the data
            try:
                g = h5.create_group(group + '/' + name)
            except ValueError:
                print("XU.io.spectra.Save2HDF5: cannot create group %s for "
                      "writing data!" % name)
                return True

            # start with saving scan comments
            for k in self.comments:
                try:
                    g.attrs[k] = self.comments[k]
                except IndexError:
                    print("XU.io.spectra.Save2HDF5: cannot save file comment "
                          "%s = %s to group %s!" % (k, self.comments[k], name))

            # save scan parameters
            for k in self.params:
                try:
                    g.attrs[k] = self.params[k]
                except IndexError:
                    print("XU.io.spectra.Save2HDF5: cannot save file parametes"
                          " %s to group %s!" % (k, name))

            # ----------finally we need to save the data -------------------
            kwds = {'fletcher32': True, 'compression': 'gzip'}

            try:
                g.create_dataset("data", data=self.data.data, **kwds)
            except (RuntimeError, ValueError):
                print("XU.io.spectra.Save2HDF5: cannot create table for "
                      "storing scan data!")
                return True

            # if there is MCA data - store this
            if self.mca is not None:
                try:
                    c = g.create_dataset(mcaname, data=self.mca, **kwds)
                except (RuntimeError, ValueError):
                    print("XU.io.spectra.Save2HDF5: cannot create carray %s "
                          "for MCA data!" % mcaname)
                    return True

                # set MCA specific attributes
                c.attrs["channels"] = self.mca_channels
                c.attrs["nchannels"] = self.mca_channels.shape[0]

            h5.flush()

        return None

    def ReadMCA(self):
        dlist = []
        for i in range(self.mca_start_index, self.mca_stop_index + 1):
            fname = self.mca_file_template % i
            data = numpy.loadtxt(fname)

            if i == self.mca_start_index:
                if len(data.shape) == 2:
                    self.mca_channels = data[:, 0]
                else:
                    self.mca_channels = numpy.arange(0, data.shape[0])

            if len(data.shape) == 2:
                dlist.append(data[:, 1].tolist())
            else:
                dlist.append(data.tolist())

        self.mca = numpy.array(dlist, dtype=float)

    def __str__(self):
        ostr = self.params.__str__()
        ostr += self.data.__str__()

        return ostr

    def Read(self):
        """
        Read the data from the file.
        """

        def addkeyval(lst, k, v):
            """
            add new key to a list. if key already exists a number will be
            appended to the key name

            Parameters
            ----------
            lst :   list
            k :     str
                key
            v :     object
                value
            """
            kcnt = 0
            key = k
            while key in lst:
                key = k + "_%i" % (kcnt + 1)
                kcnt += 1
            lst[key] = v

        col_names = []
        col_types = []
        rec_list = []
        with open(self.filename, 'rb') as fid:
            for line in fid:
                line = line.decode('utf8', 'ignore')
                line = line.strip()

                # read the next line if the line starts with a "!"
                if re_end_section.match(line):
                    continue

                # select the which section to read
                if re_comment_section.match(line):
                    read_mode = 1
                    continue

                if re_parameter_section.match(line):
                    read_mode = 2
                    continue

                if re_data_section.match(line):
                    read_mode = 3
                    continue

                # here we decide how to proceed with the data
                if read_mode == 1:
                    # read the file comments
                    try:
                        (key, value) = line.split("=")
                    except ValueError:
                        # avoid annoying output
                        if config.VERBOSITY >= config.INFO_ALL:
                            print("XU.io.SPECTRAFile.Read: cannot interpret "
                                  "the comment string: %s" % (line))
                        continue

                    key = key.strip()
                    # remove whitespaces to be conform with natural naming
                    key = key.replace(' ', '')
                    key = key.replace(':', '_')
                    # remove possible number at first position
                    if re_num.findall(key[0]) != []:
                        key = "_" + key
                    value = value.strip()
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECTRAFile.Read: comment: k, v: %s, %s"
                              % (key, value))

                    try:
                        value = float(value)
                    except ValueError:
                        pass

                    # need to handle the case, that a key may appear several
                    # times in the list
                    addkeyval(self.comments, key, value)

                elif read_mode == 2:
                    # read scan parameters
                    try:
                        (key, value) = line.split("=")
                    except ValueError:
                        print("XU.io.SPECTRAFile.Read: cannot interpret the "
                              "parameter string: %s" % (line))

                    key = key.strip()
                    # remove whitespaces to be conform with natural naming
                    key = key.replace(' ', '')
                    key = key.replace(':', '_')
                    # remove possible number at first position
                    if re_num.findall(key[0]) != []:
                        key = "_" + key
                    value = value.strip()
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECTRAFile.Read: parameter: k, v: %s, %s"
                              % (key, value))

                    try:
                        value = float(value)
                    except ValueError:
                        # if the conversion of the parameter to float
                        # fails it will be saved as a string
                        pass

                    # need to handle the case, that a key may appear several
                    # times in the list
                    addkeyval(self.params, key, value)

                elif read_mode == 3:
                    if re_column.match(line):
                        try:
                            unit = re_unit.findall(line)[0]
                        except IndexError:
                            unit = "NONE"

                        try:
                            sline = re_obracket.split(line)
                            if len(sline) == 1:
                                raise IndexError
                            lval = sline[0]
                            rval = re_cbracket.split(line)[-1]
                            dtype = rval.strip()
                            lv = re_wspaces.split(lval)
                            index = int(lv[1])
                            name = "".join(lv[2:])
                            name = name.replace(':', '_')
                        except IndexError:
                            lv = re_wspaces.split(line)
                            index = int(lv[1])
                            dtype = lv[-1]
                            name = "".join(lv[2:-1])
                            name = name.replace(':', '_')

                        # store column definition
                        self.data.append(
                            SPECTRAFileDataColumn(index, name, unit, dtype))

                        if name in col_names:
                            name += "%s_1" % name
                        col_names.append("%s" % name)
                        col_types.append("%s" % (dtype_map[dtype]))

                    else:
                        # read data
                        dlist = re_wspaces.split(line)
                        for i in range(len(dlist)):
                            dlist[i] = float(dlist[i])

                        rec_list.append(tuple(dlist))

        if config.VERBOSITY >= config.DEBUG:
            print("XU.io.SPECTRAFile.Read: data columns: name, type: %s, %s"
                  % (col_names, col_types))
        if rec_list:
            self.data.data = rec.fromrecords(rec_list, formats=col_types,
                                             names=col_names)
        else:
            self.data.data = None


def geth5_spectra_map(h5file, scans, *args, **kwargs):
    """
    function to obtain the omega and twotheta as well as intensity values
    for a reciprocal space map saved in an HDF5 file, which was created
    from a spectra file by the Save2HDF5 method.

    further more it is possible to obtain even more positions from
    the data file if more than two string arguments with its names are given

    Parameters
    ----------
    h5f :       file-handle or str
        file object of a HDF5 file opened using h5py
    scans :     int, tuple or list
        number of the scans of the reciprocal space map
    args:       str, optional
        arbitrary number of motor names

            - omname:  name of the omega motor (or its equivalent)
            - ttname:  name of the two theta motor (or its equivalent)

    kwargs :    dict, optional
    mca :       str, optional
        name of the mca data (if available) otherwise None (default: "MCA")
    samplename : str, optional
        string with the hdf5-group containing the scan data if omitted the
        first child node of h5f.root will be used to determine the sample name

    Returns
    -------
    [ang1, ang2, ...] : list
        angular positions of the center channel of the position
        sensitive detector (numpy.ndarray 1D). one entry for every
        `args`-argument given to the function
    MAP :   ndarray
        the data values as stored in the data file (includes the intensities
        e.g. MAP['MCA']).
    """

    with xu_h5open(h5file) as h5:
        mca = kwargs.get('mca', 'MCA')

        if "samplename" in kwargs:
            basename = kwargs["samplename"]
        else:
            nodename = list(h5)[0]
            basenlist = re_underscore.split(nodename)
            basename = "_".join(basenlist[:-1])
            if config.VERBOSITY >= config.DEBUG:
                print("XU.io.spectra.geth5_spectra_map: using \'%s\' as "
                      "basename" % (basename))

        if isinstance(scans, (list, tuple)):
            scanlist = scans
        else:
            scanlist = list([scans])

        angles = dict.fromkeys(args)
        for key in angles:
            angles[key] = numpy.zeros(0)
        buf = numpy.zeros(0)
        MAP = numpy.zeros(0)

        for nr in scanlist:
            h5scan = h5.get(basename + "_%05d" % nr)
            sdata = h5scan.get('data')
            if mca:
                mcanode = h5.get(basename + "_%05d/%s" % (nr, mca))
                mcadata = numpy.asarray(mcanode)

            # append scan data to MAP, where all data are stored
            mcatemp = mcadata.view([(mca, (mcadata.dtype, mcadata.shape[1]))])
            sdtmp = numpy.lib.recfunctions.merge_arrays([sdata, mcatemp],
                                                        flatten=True)
            if MAP.dtype == numpy.float64:
                MAP.dtype = sdtmp.dtype
            MAP = numpy.append(MAP, sdtmp)

            # check type of scan
            notscanmotors = []
            for i in range(len(args)):
                motname = args[i]
                try:
                    buf = sdata[motname]
                    scanshape = buf.shape
                    angles[motname] = numpy.concatenate((angles[motname], buf))
                except ValueError:
                    notscanmotors.append(i)
            for i in notscanmotors:
                motname = args[i]
                buf = numpy.ones(scanshape) * \
                    h5scan.attrs.get("%s" % motname)
                angles[motname] = numpy.concatenate((angles[motname], buf))

    retval = []
    for motname in args:
        # create return values in correct order
        retval.append(angles[motname])

    return retval, MAP



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
implement convenience functions to define Heusler materials.
"""

from . import elements
from .material import Crystal
from .spacegrouplattice import SGLattice

__all__ = ['FullHeuslerCubic225', 'FullHeuslerCubic225_A2',
           'FullHeuslerCubic225_B2', 'FullHeuslerCubic225_DO3',
           'HeuslerHexagonal194', 'HeuslerTetragonal119',
           'HeuslerTetragonal139', 'InverseHeuslerCubic216']


def _check_elements(*elem):
    ret = []
    for el in elem:
        if isinstance(el, str):
            ret.append(getattr(elements, el))
        else:
            ret.append(el)
    return ret


def FullHeuslerCubic225(X, Y, Z, a, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Full Heusler structure with formula X2YZ.
    Strukturberichte symbol L2_1; space group Fm-3m (225)

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a :         float
        cubic lattice parameter in angstrom
    biso :      list of floats, optional
        Debye Waller factors for X, Y, Z elements
    occ :       list of floats, optional
        occupation numbers for the elements X, Y, Z

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(225, a, atoms=[x, y, z], pos=['8c', '4a', '4b'],
                             b=biso, occ=occ))


def FullHeuslerCubic225_B2(X, Y, Z, a, b2dis, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Full Heusler structure with formula X2YZ.
    Strukturberichte symbol L2_1; space group Fm-3m (225) with B2-type (CsCl)
    disorder

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a :         float
        cubic lattice parameter in angstrom
    b2dis :     float
        amount of B2-type disorder (0: fully ordered, 1: fully disordered)
    biso :      list of floats, optional
        Debye Waller factors for X, Y, Z elements
    occ :       list of floats, optional
        occupation numbers for the elements X, Y, Z

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(225, a,
                             atoms=[x, y, z, y, z],
                             pos=['8c', '4a', '4b', '4b', '4a'],
                             occ=[1*occ[0], (1-b2dis/2.)*occ[1],
                                  (1-b2dis/2.)*occ[2], b2dis/2.*occ[1],
                                  b2dis/2.*occ[2]],
                             b=biso + [biso[1], biso[2]]))


def FullHeuslerCubic225_A2(X, Y, Z, a, a2dis, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Full Heusler structure with formula X2YZ.
    Strukturberichte symbol L2_1; space group Fm-3m (225) with A2-type (W)
    disorder

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a :         float
        cubic lattice parameter in angstrom
    a2dis :     float
        amount of A2-type disorder (0: fully ordered, 1: fully disordered)
    biso :      list of floats, optional
        Debye Waller factors for X, Y, Z elements
    occ :       list of floats, optional
        occupation numbers for the elements X, Y, Z

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(225, a,
                             atoms=[x, x, x, y, y, y, z, z, z],
                             pos=['8c', '4a', '4b',
                                  '8c', '4a', '4b',
                                  '8c', '4a', '4b'],
                             occ=[(1-a2dis/2.)*occ[0], a2dis/2.*occ[0],
                                  a2dis/2.*occ[0], a2dis/4.*occ[1],
                                  (1-a2dis*3./4.)*occ[1], a2dis/4.*occ[1],
                                  a2dis/4.*occ[2], a2dis/4.*occ[2],
                                  (1-a2dis*3./4.)*occ[2]],
                             b=[biso[0], ]*3 + [biso[1], ]*3 + [biso[2], ]*3))


def FullHeuslerCubic225_DO3(X, Y, Z, a, do3disxy, do3disxz, biso=[0, 0, 0],
                            occ=[1, 1, 1]):
    """
    Full Heusler structure with formula X2YZ.
    Strukturberichte symbol L2_1; space group Fm-3m (225) with DO_3-type (BiF3)
    disorder, either between atoms X <-> Y or X <-> Z.

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a :         float
        cubic lattice parameter in angstrom
    do3disxy :  float
        amount of DO_3-type disorder between X and Y atoms (0: fully ordered,
        1: fully disordered)
    do3disxz :  float
        amount of DO_3-type disorder between X and Z atoms (0: fully ordered,
        1: fully disordered)
    biso :      list of floats, optional
        Debye Waller factors for X, Y, Z elements
    occ :       list of floats, optional
        occupation numbers for the elements X, Y, Z

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(225, a,
                             atoms=[x, y, z,
                                    x, y,
                                    x, z],
                             pos=['8c', '4a', '4b',
                                  '4a', '8c',
                                  '4b', '8c'],
                             occ=[(1-do3disxy/3.-do3disxz/3.)*occ[0],
                                  (1-do3disxy*2/3.)*occ[1],
                                  (1-do3disxz*2/3.)*occ[2],
                                  do3disxy*2/3.*occ[0], do3disxy*1/3.*occ[1],
                                  do3disxz*2/3.*occ[0], do3disxz*1/3.*occ[2]],
                             b=biso + [biso[0], biso[1]] + [biso[0], biso[2]]))


def InverseHeuslerCubic216(X, Y, Z, a, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Full Heusler structure with formula (XY)X'Z structure;
    space group F-43m (216)

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a :         float
        cubic lattice parameter in angstrom

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('(%s%s)%s\'%s' % (x.basename, y.basename,
                                     x.basename, z.basename),
                   SGLattice(216, a, atoms=[x, x, y, z],
                             pos=['4a', '4d', '4b', '4c'],
                             b=[biso[0], ] + biso,
                             occ=[occ[0], ] + occ))


def HeuslerTetragonal139(X, Y, Z, a, c, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Tetragonal Heusler structure with formula X2YZ
    space group I4/mmm (139)

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a, c :      float
        tetragonal lattice parameters in angstrom

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(139, a, c,
                             atoms=[x, y, z],
                             pos=['4d', '2b', '2a'],
                             b=biso, occ=occ))


def HeuslerTetragonal119(X, Y, Z, a, c, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Tetragonal Heusler structure with formula X2YZ
    space group I-4m2 (119)

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a, c :      float
        tetragonal lattice parameters in angstrom

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s2%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(119, a, c,
                             atoms=[x, x, y, z],
                             pos=['2b', '2c', '2d', '2a'],
                             b=[biso[0], ] + biso,
                             occ=[occ[0], ] + occ))


def HeuslerHexagonal194(X, Y, Z, a, c, biso=[0, 0, 0], occ=[1, 1, 1]):
    """
    Hexagonal Heusler structure with formula XYZ
    space group P63/mmc (194)

    Parameters
    ----------
    X, Y, Z :   str or Element
        elements
    a, c :      float
        hexagonal lattice parameters in angstrom

    Returns
    -------
    Crystal
        Crystal describing the Heusler material
    """
    x, y, z = _check_elements(X, Y, Z)
    return Crystal('%s%s%s' % (x.basename, y.basename, z.basename),
                   SGLattice(194, a, c,
                             atoms=[x, y, z],
                             pos=['2a', '2c', '2d'],
                             b=biso, occ=occ))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import copy
import re
import shlex

import numpy

from .. import config
from . import xu_open

re_label = re.compile(r'^\s*_')
re_default = re.compile(r'^\s*_('
                        'pd_meas_counts_total|'
                        'pd_meas_intensity_total|'
                        'pd_proc_intensity_total|'
                        'pd_proc_intensity_net|'
                        'pd_calc_intensity_total|'
                        'pd_calc_intensity_net)')
re_loop = re.compile(r'^\s*loop_')
re_nop = re.compile(r'^\s*_(pd_meas_number_of_points|pd_meas_detector_id)')
re_multiline = re.compile(r';')


def remove_comments(line, sep='#'):
    for s in sep:
        line = line.split(s)[0]
    return line


class pdCIF(object):

    """
    the class implements a primitive parser for pdCIF-like files.  It reads
    every entry and collects the information in the header attribute. The first
    loop containing one of the intensity fields is assumed to be the data the
    user is interested in and is transfered to the data array which is stored
    as numpy record array the columns can be accessed by name

    intensity fields:

      - `_pd_meas_counts_total`
      - `_pd_meas_intensity_total`
      - `_pd_proc_intensity_total`
      - `_pd_proc_intensity_net`
      - `_pd_calc_intensity_total`
      - `_pd_calc_intensity_net`

    alternatively the data column name can be given as argument to the
    constructor
    """

    def __init__(self, filename, datacolumn=None):
        """
        contructor of the pdCIF class

        Parameters
        ----------
        filename :      str
            filename of the file to be parsed
        datacolumn :    str, optional
            name of data column to identify the data loop (default =None; means
            that a list of default names is used)
        """
        self.filename = filename
        self.datacolumn = datacolumn
        self.header = {}
        self.data = None

        self.Parse()

    def Parse(self):
        """
        parser of the pdCIF file. the method reads the data from the file and
        fills the data and header attributes with content
        """
        with xu_open(self.filename) as fh:
            self._parse_single(fh)

    def _parse_single(self, fh, breakAfterData=False):
        """
        internal routine to parse a single loop of the pdCIF file

        Parameters
        ----------
        fh :    file-handle
        breakAfterData :    bool, optional
            allowing to stop the parsing after data loop was found
            (default:False)
        """
        loopStart = False
        dataLoop = False
        dataDone = False
        loopheader = []
        numOfEntries = -1
        multiline = None
        label = None

        while True:
            line = fh.readline().decode('ascii')
            if not line:
                break

            line = remove_comments(line)
            if re_loop.match(line):
                loopStart = True
                remainingline = re.sub('loop_', '', line).strip()
                if re_label.match(remainingline):
                    if ((self.datacolumn is None and re_default.match(line)) or
                            line.strip() == self.datacolumn):
                        dataLoop = True
                    loopheader.append(remainingline)
                continue

            if multiline:
                multiline += line
                if re_multiline.match(line):  # end of multiline
                    val = multiline
                    self.header[label] = val
                    multiline = None
                continue

            if re_label.match(line) and not loopStart:
                # parse header
                split = line.split(None, 1)
                label = split[0].strip()
                try:
                    val = split[1].strip()
                    self.header[label] = val
                    # convert data format of header line
                    if re_nop.match(line):
                        numOfEntries = int(val)
                    try:
                        self.header[label] = float(val)
                    except ValueError:
                        self.header[label] = val
                except IndexError:
                    # try if multiline
                    line2 = fh.readline().decode('ascii')
                    if re_multiline.match(line2):
                        multiline = line2
                    else:  # single value must be in second line
                        self.header[label] = line2

            elif re_label.match(line) and loopStart:
                # read loop entries
                if ((self.datacolumn is None and re_default.match(line)) or
                        line.strip() == self.datacolumn):
                    dataLoop = True
                loopheader.append(line.strip())

            elif loopStart:
                fh.seek(fh.tell() - len(line))
                if numOfEntries != -1 and dataLoop and not dataDone:
                    self.data = self._parse_loop_numpy(fh, loopheader,
                                                       numOfEntries)
                    dataDone = True
                    if breakAfterData:
                        break
                elif dataLoop and not dataDone:
                    self._parse_loop(fh, loopheader)
                    length = len(self.header[loopheader[0]])
                    dtypes = [(str(entry), type(self.header[entry][0]))
                              for entry in loopheader]
                    for i in range(len(dtypes)):
                        if dtypes[i][1] is str:
                            dtypes[i] = (str(dtypes[i][0]), numpy.str_, 64)
                    self.data = numpy.zeros(length, dtype=dtypes)
                    for entry in loopheader:
                        self.data[entry] = self.header.pop(entry)
                    dataDone = True
                    if breakAfterData:
                        break
                else:
                    try:
                        self._parse_loop(fh, loopheader)
                    except ValueError:
                        if config.VERBOSITY >= config.INFO_LOW:
                            print('XU.io.pdCIF: unable to handle loop at %d'
                                  % fh.tell())
                dataLoop = False
                loopStart = False
                loopheader = []
                numOfEntries = -1

    def _parse_loop_numpy(self, filehandle, fields, nentry):
        """
        function to parse a loop using numpy routines

        Parameters
        ----------
        filehandle :    file-handle
            filehandle object to use as data source
        fields :        iterable
            field names in the loop
        nentry :        int
            number of entries in the loop

        Returns
        -------
        data :          ndarray
            data read from the file as numpy record array
        """
        tmp = numpy.fromfile(filehandle, count=nentry * len(fields), sep=' ')
        data = numpy.rec.fromarrays(tmp.reshape((-1, len(fields))).T,
                                    names=fields)
        return data

    def _parse_loop(self, filehandle, fields):
        """
        function to parse a loop using python loops routines. the fields are
        added to the fileheader dictionary

        Parameters
        ----------
        filehandle :    file-handle
            filehandle object to use as data source
        fields :        iterable
            field names in the loop

        """
        fh = filehandle

        for f in fields:
            self.header[f] = []
        while True:
            line = fh.readline().decode('ascii')
            if not line:
                break

            if re_label.match(line) or line.strip() == '':
                fh.seek(fh.tell() - len(line))
                break
            row = shlex.split(line, comments=True)
            for i in range(len(fields)):
                try:
                    self.header[fields[i]].append(float(row[i]))
                except ValueError:
                    self.header[fields[i]].append(row[i])
                except IndexError:  # maybe multiline field
                    line2 = fh.readline().decode('ascii')
                    line2 = remove_comments(line2)
                    if re_multiline.match(line2):
                        multiline = line2
                        while True:
                            line = fh.readline().decode('ascii')
                            line = remove_comments(line)
                            if not line:
                                fh.seek(fh.tell() - len(line))
                                break
                            if re_multiline.match(line) and line.strip()[1:]:
                                multiline += line
                            else:
                                self.header[fields[i]].append(multiline)
                                break
                    else:
                        fh.seek(fh.tell() - len(line2))
                        raise ValueError('a column is missing for label %s '
                                         'in a loop' % fields[i])


class pdESG(pdCIF):

    """
    class for parsing multiple pdCIF loops in one file.
    This includes especially ``*.esg`` files which are supposed to
    consist of multiple loops of pdCIF data with equal length.

    Upon parsing the class tries to combine the data of these different
    scans into a single data matrix -> same shape of subscan data is assumed
    """

    def __init__(self, filename, datacolumn=None):
        self.filename = filename
        self.datacolumn = datacolumn
        self.fileheader = {}
        self.header = {}
        self.data = None

        self.Parse()

    def Parse(self):
        """
        parser of the pdCIF file. the method reads the data from the file and
        fills the data and header attributes with content
        """
        with xu_open(self.filename) as fh:
            # parse first header and loop
            self._parse_single(fh, breakAfterData=True)
            self.fileheader = copy.deepcopy(self.header)
            self.header = {}
            fdata = self.data
            datasize = self.data.size
            nscan = 1
            tell = 0
            while True:  # try to parse all scans
                tell = fh.tell()
                self._parse_single(fh, breakAfterData=True)
                if tell == fh.tell():
                    break
                # copy changing data from header
                for key in self.header:
                    if key in self.fileheader:
                        if not isinstance(self.fileheader[key], list):
                            self.fileheader[key] = [self.fileheader[key], ]
                        self.fileheader[key].append(self.header[key])
                    else:
                        self.fileheader[key] = self.header[key]

                fdata = numpy.append(fdata, self.data)
                nscan += 1

        # convert data for output to user
        for key in self.fileheader:
            if isinstance(self.fileheader[key], list):
                self.fileheader[key] = numpy.array(self.fileheader[key])
        self.data = numpy.empty(fdata.shape)
        self.data[...] = fdata[...]
        self.data.shape = (nscan, datasize)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2017-2021 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
module handling crystal lattice structures. A SGLattice consists of a space
group number and the position of atoms specified as Wyckoff positions along
with their parameters. Depending on the space group symmetry only certain
parameters of the resulting instance will be settable! A cubic lattice for
example allows only to set its 'a' lattice parameter but none of the other unit
cell shape parameters.
"""

import copy
import fractions
import numbers
import re
from collections import OrderedDict
from math import cos, isclose, radians, sin, sqrt

import numpy
import scipy.optimize

from .. import config, cxrayutilities, math, utilities
from ..exception import InputError
from . import elements
from .atom import Atom
from .wyckpos import *

# space group number to symmetry and number of parameters dictionary
sgrp_sym = RangeDict({range(1, 3): ('triclinic', 6),
                      range(3, 16): ('monoclinic', 4),
                      range(16, 75): ('orthorhombic', 3),
                      range(75, 143): ('tetragonal', 2),
                      range(143, 168): ('trigonal', 2),
                      range(168, 195): ('hexagonal', 2),
                      range(195, 231): ('cubic', 1)})

sgrp_name = {'1': 'P1', '2': 'P-1', '3': 'P2', '4': 'P21', '5': 'C2',
             '6': 'Pm', '7': 'Pc', '8': 'Cm', '9': 'Cc', '10': 'P2/m',
             '11': 'P21/m', '12': 'C2/m', '13': 'P2/c', '14': 'P21/c',
             '15': 'C2/c', '16': 'P222', '17': 'P2221', '18': 'P21212',
             '19': 'P212121', '20': 'C2221', '21': 'C222', '22': 'F222',
             '23': 'I222', '24': 'I212121', '25': 'Pmm2', '26': 'Pmc21',
             '27': 'Pcc2', '28': 'Pma2', '29': 'Pca21', '30': 'Pnc2',
             '31': 'Pmn21', '32': 'Pba2', '33': 'Pna21', '34': 'Pnn2',
             '35': 'Cmm2', '36': 'Cmc21', '37': 'Ccc2', '38': 'Amm2',
             '39': 'Aem2', '40': 'Ama2', '41': 'Aea2', '42': 'Fmm2',
             '43': 'Fdd2', '44': 'Imm2', '45': 'Iba2', '46': 'Ima2',
             '47': 'Pmmm', '48': 'Pnnn', '49': 'Pccm', '50': 'Pban',
             '51': 'Pmma', '52': 'Pnna', '53': 'Pmna', '54': 'Pcca',
             '55': 'Pbam', '56': 'Pccn', '57': 'Pbcm', '58': 'Pnnm',
             '59': 'Pmmn', '60': 'Pbcn', '61': 'Pbca', '62': 'Pnma',
             '63': 'Cmcm', '64': 'Cmce', '65': 'Cmmm', '66': 'Cccm',
             '67': 'Cmme', '68': 'Ccce', '69': 'Fmmm', '70': 'Fddd',
             '71': 'Immm', '72': 'Ibam', '73': 'Ibca', '74': 'Imma',
             '75': 'P4', '76': 'P41', '77': 'P42', '78': 'P43',
             '79': 'I4', '80': 'I41', '81': 'P-4', '82': 'I-4',
             '83': 'P4/m', '84': 'P42/m', '85': 'P4/n', '86': 'P42/n',
             '87': 'I4/m', '88': 'I41/a', '89': 'P422', '90': 'P4212',
             '91': 'P4122', '92': 'P41212', '93': 'P4222', '94': 'P42212',
             '95': 'P4322', '96': 'P43212', '97': 'I422', '98': 'I4122',
             '99': 'P4mm', '100': 'P4bm', '101': 'P42cm', '102': 'P42nm',
             '103': 'P4cc', '104': 'P4nc', '105': 'P42mc', '106': 'P42bc',
             '107': 'I4mm', '108': 'I4cm', '109': 'I41md', '110': 'I41cd',
             '111': 'P-42m', '112': 'P-42c', '113': 'P-421m',
             '114': 'P-421c', '115': 'P-4m2', '116': 'P-4c2',
             '117': 'P-4b2', '118': 'P-4n2', '119': 'I-4m2',
             '120': 'I-4c2', '121': 'I-42m', '122': 'I-42d',
             '123': 'P4/mmm', '124': 'P4/mcc', '125': 'P4/nbm',
             '126': 'P4/nnc', '127': 'P4/mbm', '128': 'P4/mnc',
             '129': 'P4/nmm', '130': 'P4/ncc', '131': 'P42/mmc',
             '132': 'P42/mcm', '133': 'P42/nbc', '134': 'P42/nnm',
             '135': 'P42/mbc', '136': 'P42/mnm', '137': 'P42/nmc',
             '138': 'P42/ncm', '139': 'I4/mmm', '140': 'I4/mcm',
             '141': 'I41/amd', '142': 'I41/acd', '143': 'P3',
             '144': 'P31', '145': 'P32', '146': 'R3', '147': 'P-3',
             '148': 'R-3', '149': 'P312', '150': 'P321', '151': 'P3112',
             '152': 'P3121', '153': 'P3212', '154': 'P3221', '155': 'R32',
             '156': 'P3m1', '157': 'P31m', '158': 'P3c1', '159': 'P31c',
             '160': 'R3m', '161': 'R3c', '162': 'P-31m', '163': 'P-31c',
             '164': 'P-3m1', '165': 'P-3c1', '166': 'R-3m', '167': 'R-3c',
             '168': 'P6', '169': 'P61', '170': 'P65', '171': 'P62',
             '172': 'P64', '173': 'P63', '174': 'P-6', '175': 'P6/m',
             '176': 'P63/m', '177': 'P622', '178': 'P6122',
             '179': 'P6522', '180': 'P6222', '181': 'P6422',
             '182': 'P6322', '183': 'P6mm', '184': 'P6cc', '185': 'P63cm',
             '186': 'P63mc', '187': 'P-6m2', '188': 'P-6c2',
             '189': 'P-62m', '190': 'P-62c', '191': 'P6/mmm',
             '192': 'P6/mcc', '193': 'P63/mcm', '194': 'P63/mmc',
             '195': 'P23', '196': 'F23', '197': 'I23', '198': 'P213',
             '199': 'I213', '200': 'Pm-3', '201': 'Pn-3', '202': 'Fm-3',
             '203': 'Fd-3', '204': 'Im-3', '205': 'Pa-3', '206': 'Ia-3',
             '207': 'P432', '208': 'P4232', '209': 'F432',
             '210': 'F4132', '211': 'I432', '212': 'P4332',
             '213': 'P4132', '214': 'I4132', '215': 'P-43m',
             '216': 'F-43m', '217': 'I-43m', '218': 'P-43n',
             '219': 'F-43c', '220': 'I-43d', '221': 'Pm-3m',
             '222': 'Pn-3n', '223': 'Pm-3n', '224': 'Pn-3m',
             '225': 'Fm-3m', '226': 'Fm-3c', '227': 'Fd-3m',
             '228': 'Fd-3c', '229': 'Im-3m', '230': 'Ia-3d'}

sgrp_params = {'cubic:1': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'cubic:2': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'cubic': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'hexagonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'trigonal:R': (('a', 'alpha'), ('a', 'a', 'a', 'alpha',
                                               'alpha', 'alpha')),
               'trigonal:H': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'trigonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'tetragonal:1': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'tetragonal:2': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'tetragonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'orthorhombic:1': (('a', 'b', 'c'),
                                  ('a', 'b', 'c', 90, 90, 90)),
               'orthorhombic:2': (('a', 'b', 'c'),
                                  ('a', 'b', 'c', 90, 90, 90)),
               'orthorhombic': (('a', 'b', 'c'),
                                ('a', 'b', 'c', 90, 90, 90)),
               'monoclinic:b': (('a', 'b', 'c', 'beta'),
                                ('a', 'b', 'c', 90, 'beta', 90)),
               'monoclinic:c': (('a', 'b', 'c', 'gamma'),
                                ('a', 'b', 'c', 90, 90, 'gamma')),
               'monoclinic': (('a', 'b', 'c', 'beta'),
                              ('a', 'b', 'c', 90, 'beta', 90)),
               'triclinic': (('a', 'b', 'c', 'alpha', 'beta', 'gamma'),
                             ('a', 'b', 'c', 'alpha', 'beta', 'gamma'))}

# regular expression for splitting multiple reflection conditions
hklcond_group = re.compile(r'([-hkil0-9\(\)]+): ([-+hklnor1-8=\s,]+)(?:, |$)')


def get_possible_sgrp_suf(sgrp_nr):
    """
    determine possible space group suffix. Multiple suffixes might be possible
    for one space group due to different origin choice, unique axis, or choice
    of the unit cell shape.

    Parameters
    ----------
    sgrp_nr :   int
        space group number

    Returns
    -------
    str or list
        either an empty string or a list of possible valid suffix strings
    """
    sgrp_suf = ''
    if sgrp_nr in [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]:
        sgrp_suf = [':b', ':c']
    elif sgrp_nr in [48, 50, 59, 68, 70, 85, 86, 88, 125, 126,
                     129, 130, 133, 134, 137, 138, 141, 142,
                     201, 203, 222, 224, 227, 228]:
        sgrp_suf = [':1', ':2']
    elif sgrp_nr in [146, 148, 155, 160, 161, 166, 167]:
        sgrp_suf = [':H', ':R']
    return sgrp_suf


def get_default_sgrp_suf(sgrp_nr):
    """
    determine default space group suffix
    """
    possibilities = get_possible_sgrp_suf(sgrp_nr)
    if possibilities:
        return possibilities[0]
    else:
        return ''


def _get_pardict(parint, x):
    """
    internal helper function to determine an parameter dictionary needed for
    the evaluation of Wyckoff positions

    Parameters
    ----------
    parint :    int
        integer number specifying the parameters (binary code, see
        implementation)
    x :         tuple, list
        parameter values in an apropriate length list or tuple

    Returns
    -------
    dict
    """
    i = 0
    pardict = {}
    if parint & 1:
        pardict['x'] = x[i]
        i += 1
    if parint & 2:
        pardict['y'] = x[i]
        i += 1
    if parint & 4:
        pardict['z'] = x[i]
    return pardict


def testwp(parint, wp, cifpos, digits=config.DIGITS):
    """
    test if a Wyckoff position can describe the given position from a CIF file

    Parameters
    ----------
    parint :    int
        telling which Parameters the given Wyckoff position has
    wp :        str or tuple
        expression of the Wyckoff position
    cifpos :    list, or tuple or array-like
        (x, y, z) position of the atom in the CIF file
    digits :    int
        number of digits for which for a comparison of floating point numbers
        will be rounded to. By default xu.config.DIGITS is used.

    Returns
    -------
    foundflag :     bool
        flag to tell if the positions match
    pars :          array-like or None
        parameters associated with the position or None if no parameters are
        needed
    """
    def check_positions_match(p1, p2, digits):
        p1 = p1 - numpy.round(p1, digits) // 1
        p2 = p2 - numpy.round(p2, digits) // 1
        if numpy.round(p1, digits) == numpy.round(p2, digits):
            return True
        else:
            return False

    wyckp = wp.strip('()').split(',')
    # test agreement in positions witout variables
    match = [False, False, False]
    variables = []
    for i in range(3):
        v = re.findall(r'[xyz]', wyckp[i])
        if v == []:
            pos = eval(wyckp[i])
            match[i] = check_positions_match(pos, cifpos[i], digits)
            if not match[i]:
                return False, None
        else:
            variables += v

    if all(match):
        return True, None

    # check if with proper choice of the variables a correspondence of the
    # positions can be obtained
    def fmin(x, parint, wyckp, cifpos):
        evalexp = []
        cifp = []
        for i in range(3):
            if not match[i]:
                evalexp.append(wyckp[i])
                cifp.append(cifpos[i])
        pardict = _get_pardict(parint, x)
        wpos = [eval(e, pardict) for e in evalexp]
        return numpy.linalg.norm(numpy.subtract(wpos, cifp))

    x0 = []
    if 'x' in variables:
        x0.append(cifpos[0])
    if 'y' in variables:
        x0.append(cifpos[1])
    if 'z' in variables:
        x0.append(cifpos[2])

    opt = scipy.optimize.minimize(fmin, x0, args=(parint, wyckp, cifpos))
    pardict = _get_pardict(parint, opt.x)
    for i in range(3):
        if not match[i]:
            pos = eval(wyckp[i], pardict)
            match[i] = check_positions_match(pos, cifpos[i], digits)
    if all(match):
        return True, list(opt.x)
    else:
        return False, None


def get_wyckpos(sgrp, atompos):
    """
    test all Wyckoff positions on every atomic position

    Parameters
    ----------
    sgrp : str
        space group name
    atompos : list
        list of atomic positions to identify. All atomic positions are expected
        to belong to one and the same Wyckoff position!

    Returns
    -------
    position argument for WyckoffBase.append
    """
    for k, wyckpos in wp[sgrp].items():
        parint, poslist, reflcond = wyckpos
        item = poslist[0]
        for pos in atompos:
            foundwp, par = testwp(parint, item, pos)
            if foundwp:
                return k if par is None else (k, list(par))


class WyckoffBase(list):

    """
    The WyckoffBase class implements a container for a set of Wyckoff positions
    that form the base of a crystal lattice. An instance of this class can be
    treated as a simple container object.
    """

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)

    @staticmethod
    def _checkatom(atom):
        if isinstance(atom, str):
            atom = getattr(elements, atom)
        elif not isinstance(atom, Atom):
            raise TypeError("atom must be an instance of class "
                            "xrayutilities.materials.Atom")
        return atom

    @staticmethod
    def _checkpos(pos):
        if isinstance(pos, str):
            pos = (pos, None)
        elif isinstance(pos, (tuple, list)):
            if len(pos) == 1:
                pos = (pos[0], None)
            elif len(pos) == 2:
                if isinstance(pos[1], numbers.Number):
                    pos = (pos[0], (pos[1], ))
                elif pos[1] is None:
                    pos = (pos[0], None)
                else:
                    pos = (pos[0], tuple(pos[1]))
            else:
                if isinstance(pos[1], numbers.Number):
                    pos = (pos[0], tuple(pos[1:]))
        return pos

    def append(self, atom, pos, occ=1.0, b=0.):
        """
        add new Atom to the lattice base

        Parameters
        ----------
        atom :   Atom
            object to be added
        pos :    tuple or str
            Wyckoff position of the atom, along with its parameters.
            Examples: ('2i', (0.1, 0.2, 0.3)), or '1a'
        occ :    float, optional
            occupancy (default=1.0)
        b :      float, optional
            b-factor of the atom used as exp(-b*q**2/(4*pi)**2) to reduce the
            intensity of this atom (only used in case of temp=0 in
            StructureFactor and chi calculation)
        """
        atom = self._checkatom(atom)
        pos = self._checkpos(pos)
        list.append(self, (atom, pos, occ, b))

    def __setitem__(self, key, data):
        (atom, pos, occ, b) = data
        atom = self._checkatom(atom)
        pos = self._checkpos(pos)
        list.__setitem__(self, key, (atom, pos, float(occ), float(b)))

    def __copy__(self):
        """
        since we use a custom 'append' method we need to overwrite copy
        """
        cls = self.__class__
        new = cls.__new__(cls)
        for item in self:
            new.append(*item)
        return new

    def __deepcopy__(self, memo):
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new
        for item in self:
            citem = copy.deepcopy(item, memo)
            new.append(*citem)
        return new

    def __str__(self):
        ostr = ''
        for i, (atom, p, occ, b) in enumerate(self):
            ostr += '%d: %s %s ' % (i, str(atom), p[0])
            if p[1] is not None:
                ostr += ' '.join(map(utilities.frac2str, p[1]))
            ostr += ' occ=%5.3f b=%5.3f\n' % (occ, b)
        return ostr

    def __contains__(self, item):
        """
        check if this list contains already the same element, at the same
        position, and with the same Debye Waller factor. The occupancy is not
        checked intentionally.

        Parameters
        ----------
        item :  tuple or list
            WyckoffBase entry to check if its present in this list

        Returns
        -------
        bool
        """
        for atom, p, occ, b in self:
            if (atom == item[0] and self.pos_eq(p, item[1]) and
                    isclose(b, item[3], abs_tol=1e-4)):
                return True
        return False

    @staticmethod
    def entry_eq(e1, e2):
        """
        compare two entries including all its properties to be equal

        Parameters
        ----------
        e1, e2:     tuple
            tuples with length 4 containing the entries of WyckoffBase which
            should be compared
        """
        if (e1[0] == e2[0] and WyckoffBase.pos_eq(e1[1], e2[1]) and
                numpy.allclose(e1[2:], e2[2:], atol=1e-4)):
            return True
        return False

    @staticmethod
    def pos_eq(pos1, pos2):
        """
        compare Wyckoff positions

        Parameters
        ----------
        pos1, pos2:     tuple
            tuples with Wyckoff label and optional parameters
        """
        if pos1[0] != pos2[0]:
            return False
        if pos1[1] == pos2[1]:
            return True
        else:
            for f1, f2 in zip(pos1[1], pos2[1]):
                if not isclose(f1 % 1, f2 % 1, abs_tol=1e-5):
                    return False
        return True

    def index(self, item):
        """
        return the index of the atom (same element, position, and Debye Waller
        factor). The occupancy is not checked intentionally. If the item is not
        present a ValueError is raised.

        Parameters
        ----------
        item :  tuple or list
            WyckoffBase entry

        Returns
        -------
        int
        """
        for i, (atom, p, occ, b) in enumerate(self):
            if (atom == item[0] and self.pos_eq(p, item[1]) and
                    isclose(b, item[3], abs_tol=1e-4)):
                return i
        raise ValueError("%s is not in list" % str(item))


class SymOp(object):
    """
    Class descriping a symmetry operation in a crystal. The symmetry operation
    is characterized by a 3x3 transformation matrix as well as a 3-vector
    describing a translation. For magnetic symmetry operations also the time
    reversal symmetry can be specified (not used in xrayutilities)
    """

    def __init__(self, D, t, m=1):
        """
        Initialize the symmetry operation

        Parameters
        ----------
        D : array-like
            transformation matrix (3x3)
        t : array-like
            translation vector (3)
        m : int, optional
            indicates time reversal in magnetic groups. +1 (default, no time
            reveral) or -1
        """
        self._W = numpy.zeros((4, 4))
        self._W[:3, :3] = numpy.asarray(D)
        self._W[:3, 3] = numpy.asarray(t)
        self._W[3, 3] = 1
        self._m = m

    @classmethod
    def from_xyz(cls, xyz):
        """
        create a SymOp from the xyz notation typically used in CIF files.

        Parameters
        ----------
         xyz :   str
            string describing the symmetry operation (e.g. '-y, -x, z')
        """
        D = numpy.zeros((3, 3))
        t = numpy.array(eval(xyz, {'x': 0, 'y': 0, 'z': 0})[:3])
        m = 1
        for i, expr in enumerate(xyz.strip('()').split(',')):
            if i == 3:  # time reversal property
                m = int(expr)
                continue
            if 'x' in expr:
                D[i, 0] = -1 if '-x' in expr else 1
            if 'y' in expr:
                D[i, 1] = -1 if '-y' in expr else 1
            if 'z' in expr:
                D[i, 2] = -1 if '-z' in expr else 1
        return SymOp(D, t, m)

    def xyz(self, showtimerev=False):
        """
        return the symmetry operation in xyz notation
        """
        ret = ''
        t = self.t
        for i in range(3):
            expr = ''
            if abs(self._W[i, 0]) == 1:
                expr += '+x' if self._W[i, 0] == 1 else '-x'
            if abs(self._W[i, 1]) == 1:
                expr += '+y' if self._W[i, 1] == 1 else '-y'
            if abs(self._W[i, 2]) == 1:
                expr += '+z' if self._W[i, 2] == 1 else '-z'
            if t[i] != 0:
                expr += '+' if t[i] > 0 else ''
                expr += str(fractions.Fraction(t[i]).limit_denominator(100))
            expr = expr.strip('+')
            ret += expr + ', '
        if showtimerev:
            ret += '{:+d}'.format(self._m)
        return ret.strip(', ')

    @property
    def D(self):
        """transformation matrix of the symmetry operation"""
        return self._W[:3, :3]

    @property
    def t(self):
        """translation vector of the symmetry operation"""
        return self._W[:3, 3]

    def __eq__(self, other):
        if not isinstance(other, SymOp):
            return NotImplemented
        return self._m == other._m and numpy.all(self._W == other._W)

    @staticmethod
    def foldback(v):
        return v - numpy.round(v, config.DIGITS) // 1

    def apply_rotation(self, vec):
        return self.D @ vec

    def apply(self, vec, foldback=True):
        lv = numpy.asarray(list(vec) + [1, ])
        result = (self._W @ lv)[:3]
        if foldback:
            return self.foldback(result)
        return result

    def apply_axial(self, vec):
        return self._m * numpy.linalg.det(self.D) * self.D @ vec

    def combine(self, other):
        if not isinstance(other, SymOp):
            return NotImplemented
        W = self._W @ other._W
        return SymOp(W[:3, :3], self.foldback(W[:3, 3]), self._m*other._m)

    def __str__(self):
        return '({})'.format(self.xyz(showtimerev=True))

    def __repr__(self):
        return self.__str__()


class SGLattice(object):
    """
    lattice object created from the space group number and corresponding unit
    cell parameters. atoms in the unit cell are specified by their Wyckoff
    position and their free parameters.
    """

    def __init__(self, sgrp, *args, **kwargs):
        """
        initialize class with space group number and atom list

        Parameters
        ----------
        sgrp :  int or str
            Space group number
        *args : float
            space group parameters. depending on the space group number this
            are 1 (cubic) to 6 (triclinic) parameters.
            cubic : a (lattice parameter).
            hexagonal : a, c.
            trigonal : a, c.
            tetragonal : a, c.
            orthorhombic : a, b, c.
            monoclinic : a, b, c, beta (in degree).
            triclinic : a, b, c, alpha, beta, gamma (in degree).
        atoms :     list, optional
            list of elements either as Element object or string with the
            element name. If you specify atoms you have to also give the same
            number of Wyckoff positions
        pos :       list, optional
            list of the atomic positions within the unit cell. This can be
            given as Wyckoff position along with its parameters or any position
            of an atom which will be used to identify the Wyckoff position. If
            a position has no free parameter the parameters can be omitted.
            Example: [('2i', (0.1, 0.2, 0.3)), '1a', (0, 0.5, 0)]
        occ :       list, optional
            site occupation for the atoms. This is optional and defaults to 1
            if not given.
        b :         list, optional
            b-factor of the atom used as exp(-b*q**2/(4*pi)**2) to reduce the
            intensity of this atom (only used in case of temp=0 in
            StructureFactor and chi calculation)
        """
        valid_kwargs = {'atoms': 'list of elements',
                        'pos': 'list of Wyckoff positions',
                        'occ': 'site occupations',
                        'b': 'Debye Waller exponents'}
        utilities.check_kwargs(kwargs, valid_kwargs, self.__class__.__name__)
        self.space_group = str(sgrp)
        self.space_group_nr = int(self.space_group.split(':')[0])
        try:
            self.space_group_suf = ':' + self.space_group.split(':')[1]
        except IndexError:
            self.space_group_suf = get_default_sgrp_suf(self.space_group_nr)

        if self.space_group_suf != '':
            self.space_group = str(self.space_group_nr) + self.space_group_suf
        self.name = sgrp_name[str(self.space_group_nr)] + self.space_group_suf
        self.crystal_system, nargs = sgrp_sym[self.space_group_nr]
        self.crystal_system += self.space_group_suf
        if len(args) != nargs:
            raise ValueError('XU: number of parameters (%d) does not match the'
                             ' crystal symmetry (%s:%d)'
                             % (len(args), self.crystal_system, nargs))
        self.free_parameters = OrderedDict()
        for a, par in zip(args, sgrp_params[self.crystal_system][0]):
            self.free_parameters[par] = a

        self._parameters = OrderedDict()
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                self._parameters[p] = self.free_parameters[key]
            else:
                self._parameters[p] = key

        # define lattice vectors
        self._ai = numpy.zeros((3, 3))
        self._bi = numpy.empty((3, 3))
        a, b, c, alpha, beta, gamma = self._parameters.values()
        ra = radians(alpha)
        self._paramhelp = [cos(ra), cos(radians(beta)),
                           cos(radians(gamma)), sin(ra), 0]
        self._setlat()
        # save general Wyckoff position
        self._gplabel = sorted(wp[self.space_group],
                               key=lambda s: int(s[:-1]))[-1]
        self._gp = wp[self.space_group][self._gplabel]

        # set atom positions in the lattice base
        self._wbase = WyckoffBase()
        atoms = kwargs.get('atoms', None)
        wps = kwargs.get('pos', None)
        if atoms:
            occs = kwargs.get('occ', [1.0, ] * len(atoms))
            bs = kwargs.get('b', [0.0, ] * len(atoms))
            for at, wpos, o, b in zip(atoms, wps, occs, bs):
                if (not isinstance(wpos, (tuple, list, numpy.ndarray)) or
                        len(wpos) < 3):
                    self._wbase.append(at, wpos, o, b)
                else:
                    # atomic position given -> identify Wyckoff position
                    # find all equivalent positions
                    gplist = set()
                    for p in self._gp[1]:
                        pos = eval(p,
                                   {'x': wpos[0], 'y': wpos[1], 'z': wpos[2]})
                        pos = SymOp.foldback(pos)
                        gplist.add(tuple(pos))

                    wyckpos = get_wyckpos(self.space_group, gplist)
                    if config.VERBOSITY >= config.INFO_LOW:
                        print(f"XU.materials.SGLattice: position {wpos} "
                              f"identified as {wyckpos}")

                    self._wbase.append(at, wyckpos, o, b)
        self.nsites = len(self._wbase)

        # symmetry operations and reflection conditions placeholder
        self._hklmat = []
        self._symops = []
        self._hklcond = []
        self._hklcond_wp = []
        self._iscentrosymmetric = None

    @property
    def symops(self):
        """
        return the set of symmetry operations from the general Wyckoff
        position of the space group.
        """
        if self._symops == []:
            for p in self._gp[1]:
                self._symops.append(SymOp.from_xyz(p))
        return self._symops

    @property
    def _hklsym(self):
        if self._hklmat == []:
            for s in self.symops:
                self._hklmat.append(numpy.round(self._qtransform.imatrix @
                                                self._transform.matrix @ s.D @
                                                self._transform.imatrix @
                                                self._qtransform.matrix,
                                                config.DIGITS))
        return self._hklmat

    def base(self):
        """
        generator of atomic position within the unit cell.
        """
        if not self._wbase:
            return
        sgwp = wp[self.space_group]
        for (atom, w, occ, b) in self._wbase:
            x, y, z = None, None, None
            parint, poslist, dummy = sgwp[w[0]]
            i = 0
            if parint & 1:
                try:
                    x = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if parint & 2:
                try:
                    y = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if parint & 4:
                try:
                    z = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if w[1]:
                if i != len(w[1]):
                    raise TypeError('XU.materials: too many parameters for '
                                    'Wyckoff position')

            for p in poslist:
                pos = eval(p, {'x': x, 'y': y, 'z': z})
                pos = SymOp.foldback(pos)
                yield atom, pos, occ, b

    def _setlat(self):
        a, b, c, alpha, beta, gamma = self._parameters.values()
        ca, cb, cg, sa, vh = self._paramhelp
        vh = sqrt(1 - ca**2-cb**2-cg**2 + 2*ca*cb*cg)
        self._paramhelp[4] = vh
        self._ai[0, 0] = a * vh / sa
        self._ai[0, 1] = a * (cg-cb*ca) / sa
        self._ai[0, 2] = a * cb
        self._ai[1, 1] = b * sa
        self._ai[1, 2] = b * ca
        self._ai[2, 2] = c
        self._transform = math.Transform(self._ai.T)
        self._setb()

    def _setb(self):
        V = self.UnitCellVolume()
        p = 2. * numpy.pi / V
        math.VecCross(p*self._ai[1, :], self._ai[2, :], out=self._bi[0, :])
        math.VecCross(p*self._ai[2, :], self._ai[0, :], out=self._bi[1, :])
        math.VecCross(p*self._ai[0, :], self._ai[1, :], out=self._bi[2, :])
        self._qtransform = math.Transform(self._bi.T)

    def _set_params_from_sym(self):
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                if p not in self.free_parameters:
                    self._parameters[p] = self.free_parameters[key]

    @property
    def a(self):
        return self._parameters['a']

    @a.setter
    def a(self, value):
        if 'a' not in self.free_parameters:
            raise RuntimeError("a can not be set, its not a free parameter!")
        self._parameters['a'] = value
        self.free_parameters['a'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def b(self):
        return self._parameters['b']

    @b.setter
    def b(self, value):
        if 'b' not in self.free_parameters:
            raise RuntimeError("b can not be set, its not a free parameter!")
        self._parameters['b'] = value
        self.free_parameters['b'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def c(self):
        return self._parameters['c']

    @c.setter
    def c(self, value):
        if 'c' not in self.free_parameters:
            raise RuntimeError("c can not be set, its not a free parameter!")
        self._parameters['c'] = value
        self.free_parameters['c'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def alpha(self):
        return self._parameters['alpha']

    @alpha.setter
    def alpha(self, value):
        if 'alpha' not in self.free_parameters:
            raise RuntimeError("alpha can not be set for this space group!")
        self._parameters['alpha'] = value
        self.free_parameters['alpha'] = value
        self._set_params_from_sym()
        ra = radians(value)
        self._paramhelp[0] = cos(ra)
        self._paramhelp[3] = sin(ra)
        self._setlat()

    @property
    def beta(self):
        return self._parameters['beta']

    @beta.setter
    def beta(self, value):
        if 'beta' not in self.free_parameters:
            raise RuntimeError("beta can not be set for this space group!")
        self._parameters['beta'] = value
        self.free_parameters['beta'] = value
        self._set_params_from_sym()
        self._paramhelp[1] = cos(radians(value))
        self._setlat()

    @property
    def gamma(self):
        return self._parameters['gamma']

    @gamma.setter
    def gamma(self, value):
        if 'gamma' not in self.free_parameters:
            raise RuntimeError("gamma can not be set for this space group!")
        self._parameters['gamma'] = value
        self.free_parameters['gamma'] = value
        self._set_params_from_sym()
        self._paramhelp[2] = cos(radians(value))
        self._setlat()

    def __eq__(self, other):
        """
        compare another SGLattice instance to decide if both are equal.
        To be equal they have to use the same space group, have equal lattice
        parameters and contain equal atoms in their base.
        """
        if self.space_group != other.space_group:
            return False
        # compare lattice parameters
        for prop in self.free_parameters:
            if not isclose(getattr(self, prop), getattr(other, prop)):
                return False
        # compare atoms in base
        for e in self._wbase:
            if e not in other._wbase:
                return False
            idx = other._wbase.index(e)
            if not WyckoffBase.entry_eq(e, other._wbase[idx]):
                return False
        return True

    def GetPoint(self, *args):
        """
        determine lattice points with indices given in the argument

        Examples
        --------
        >>> xu.materials.Si.lattice.GetPoint(0, 0, 4)
        array([  0.     ,   0.     ,  21.72416])

        or

        >>> xu.materials.Si.lattice.GetPoint((1, 1, 1))
        array([ 5.43104,  5.43104,  5.43104])
        """
        if len(args) == 1:
            args = args[0]
        return self._transform(args)

    def GetQ(self, *args):
        """
        determine the reciprocal lattice points with indices given in the
        argument
        """
        if len(args) == 1:
            args = args[0]
        return self._qtransform(args)

    def GetHKL(self, *args):
        """
        determine the Miller indices of the given reciprocal lattice points
        """
        if len(args) == 1:
            args = args[0]
        return self._qtransform.inverse(args)

    def UnitCellVolume(self):
        """
        function to calculate the unit cell volume of a lattice (angstrom^3)
        """
        a, b, c, alpha, beta, gamma = self._parameters.values()
        return a * b * c * self._paramhelp[4]

    def ApplyStrain(self, eps):
        """
        Applies a certain strain on a lattice. The result is a change
        in the base vectors. The full strain matrix (3x3) needs to be given.

        Note:
            Here you specify the strain and not the stress -> NO elastic
            response of the material will be considered!

        Note:
            Although the symmetry of the crystal can be lowered by this
            operation the spacegroup remains unchanged! The 'free_parameters'
            attribute is, however, updated to mimic the possible reduction of
            the symmetry.

        Parameters
        ----------
        eps :    array-like
            a 3x3 matrix with all strain components
        """
        if isinstance(eps, (list, tuple)):
            eps = numpy.asarray(eps, dtype=numpy.double)
        if eps.shape != (3, 3):
            raise InputError("ApplyStrain needs a 3x3 matrix "
                             "with strain values")

        ai = self._ai + numpy.dot(eps, self._ai.T).T
        self._parameters['a'] = math.VecNorm(ai[0, :])
        self._parameters['b'] = math.VecNorm(ai[1, :])
        self._parameters['c'] = math.VecNorm(ai[2, :])
        self._parameters['alpha'] = math.VecAngle(ai[1, :], ai[2, :], deg=True)
        self._parameters['beta'] = math.VecAngle(ai[0, :], ai[2, :], deg=True)
        self._parameters['gamma'] = math.VecAngle(ai[0, :], ai[1, :], deg=True)
        # update helper parameters
        ra = radians(self._parameters['alpha'])
        self._paramhelp[0] = cos(ra)
        self._paramhelp[1] = cos(radians(self._parameters['beta']))
        self._paramhelp[2] = cos(radians(self._parameters['gamma']))
        self._paramhelp[3] = sin(ra)
        # set new transformations
        self._setlat()
        # update free_parameters
        for p, v in self.free_parameters.items():
            self.free_parameters[p] = self._parameters[p]
        # artificially reduce symmetry if needed
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                if self._parameters[p] != self.free_parameters[key]:
                    self.free_parameters[p] = self._parameters[p]
            else:
                if self._parameters[p] != key:
                    self.free_parameters[p] = self._parameters[p]

    @property
    def iscentrosymmetric(self):
        """
        returns a boolean to determine if the lattice has centrosymmetry.
        """
        if self._iscentrosymmetric is None:
            self._iscentrosymmetric = False
            for s in self.symops:
                if numpy.all(-numpy.identity(3) == s.D):
                    self._iscentrosymmetric = True
                    break
        return self._iscentrosymmetric

    def isequivalent(self, hkl1, hkl2):
        """
        determining if hkl1 and hkl2 are two crystallographical equivalent
        pairs of Miller indices. Note that this function considers the effect
        of non-centrosymmetry!

        Parameters
        ----------
        hkl1, hkl2 :    list
            Miller indices to be checked for equivalence

        Returns
        -------
        bool
        """
        return tuple(hkl2) in self.equivalent_hkls(hkl1)

    def equivalent_hkls(self, hkl):
        """
        returns a list of equivalent hkl peaks depending on the crystal system
        """
        suf = self.space_group_suf
        nr = self.space_group_nr
        if suf == get_default_sgrp_suf(nr):
            ehkl = set(eqhkl_default[nr](hkl[0], hkl[1], hkl[2]))
        elif suf in get_possible_sgrp_suf(nr):
            ehkl = set(eqhkl_custom[nr](hkl[0], hkl[1], hkl[2]))
        else:  # fallback calculation with symmetry operations
            ehkl = numpy.unique(numpy.einsum('...ij,j', self._hklsym, hkl),
                                axis=0)
            ehkl = set(tuple(e) for e in ehkl)
        return ehkl

    def hkl_allowed(self, hkl, returnequivalents=False):
        """
        check if Bragg reflection with Miller indices hkl can exist according
        to the reflection conditions. If no reflection conditions are available
        this function returns True for all hkl values!

        Parameters
        ----------
        hkl : tuple or list
         Miller indices of the reflection to check
        returnequivalents : bool, optional
         If True all the equivalent Miller indices of hkl are returned in a
         set as second return argument.

        Returns
        -------
        allowed : bool
         True if reflection can have non-zero structure factor, false otherwise
        equivalents : set, optional
         set of equivalent Miller indices if returnequivalents is True
        """
        # generate all equivalent hkl values which also need to be checked:
        hkls = self.equivalent_hkls(hkl)

        def build_return(allowed, requi=returnequivalents):
            if requi:
                return allowed, hkls
            else:
                return allowed

        # load reflection conditions if needed
        if self._gp[2] == 'n/a':
            return build_return(True)
        if self._hklcond == [] and self._gp[2] is not None:
            self._hklcond = hklcond_group.findall(self._gp[2])
        if self._hklcond_wp == []:
            for lab in set([e[1][0] for e in self._wbase]):
                if lab == self._gplabel:  # if gen. pos. is occupied skip it
                    self._hklcond_wp.append(None)
                elif wp[self.space_group][lab][2] is None:
                    self._hklcond_wp.append(None)
                else:
                    self._hklcond_wp.append(hklcond_group.findall(
                        wp[self.space_group][lab][2]))

        # call C-code to (efficiently) test the conditions
        ret = cxrayutilities.testhklcond(hkls, self._hklcond, self._hklcond_wp)
        return build_return(ret)

    def get_allowed_hkl(self, qmax):
        """
        return a set of all allowed reflections up to a maximal specified
        momentum transfer.

        Parameters
        ----------
        qmax : float
         maximal momentum transfer

        Returns
        -------
        hklset : set
         set of allowed hkl reflections
        """
        def recurse_hkl(h, k, l, kstep):
            if (h, k, l) in hkltested:
                return
            m = self._qtransform.matrix
            q = m[:, 0]*h + m[:, 1]*k + m[:, 2]*l  # efficient matmul
            if sqrt(q[0]**2 + q[1]**2 + q[2]**2) >= qmax:
                return
            else:
                allowed, eqhkl = self.hkl_allowed((h, k, l),
                                                  returnequivalents=True)
                hkltested.update(eqhkl)
                if not self.iscentrosymmetric:
                    hkltested.update((-h, -k, -l) for (h, k, l) in eqhkl)
                if allowed:
                    hklset.update(eqhkl)
                    if not self.iscentrosymmetric:
                        eqhkl = self.equivalent_hkls((-h, -k, -l))
                        hklset.update(eqhkl)
                recurse_hkl(h+1, k, l, kstep)
                recurse_hkl(h, k+kstep, l, kstep)
                recurse_hkl(h, k, l+1, kstep)
                recurse_hkl(h, k, l-1, kstep)

        hklset = set()
        hkltested = set()
        q = numpy.empty(3)
        recurse_hkl(0, 0, 0, +1)
        recurse_hkl(1, -1, 0, -1)
        hklset.remove((0, 0, 0))
        return hklset

    def reflection_conditions(self):
        """
        return string of reflection conditions, both general (from space group)
        and of Wyckoff positions
        """
        ostr = "Reflection conditions:\n"
        ostr += " general: %s\n" % str(self._gp[2])
        for wplabel in set([e[1][0] for e in self._wbase]):
            ostr += "%8s: %s \n" % (wplabel,
                                    str(wp[self.space_group][wplabel][2]))
        return ostr

    def __str__(self):
        ostr = "{sg} {cs} {n}: a = {a:.4f}, b = {b:.4f} c= {c:.4f}\n" +\
               "alpha = {alpha:.3f}, beta = {beta:.3f}, gamma = {gamma:.3f}\n"
        ostr = ostr.format(sg=self.space_group, cs=self.crystal_system,
                           n=self.name, **self._parameters)
        if self._wbase:
            ostr += "Lattice base:\n"
            ostr += str(self._wbase)
        ostr += self.reflection_conditions()
        return ostr

    def convert_to_P1(self):
        """
        create a P1 equivalent of this SGLattice instance.

        Returns
        -------
        SGLattice
            instance with the same properties as the present lattice, however,
            in the P1 setting.
        """
        a, b, c, alpha, beta, gamma = (self.a, self.b, self.c, self.alpha,
                                       self.beta, self.gamma)
        atoms = []
        pos = []
        occ = []
        biso = []
        for at, p, o, bf in self.base():
            atoms.append(at)
            pos.append(('1a', p))
            occ.append(o)
            biso.append(bf)
        return type(self)(1, a, b, c, alpha, beta, gamma, atoms=atoms, pos=pos,
                          occ=occ, b=biso)

    def findsym(self):
        """
        method to return the highest symmetry description of the current
        material. This method does not consider to change the unit cell
        dimensions but only searches the highest symmetry spacegroup which with
        the current unit cell setting can be described. It is therefore not an
        implementation of FINDSYM [1].

        Returns
        -------
        new SGLattice-instance
            a new SGLattice instance is returned with the highest available
            symmetry description. (see restrictions above)

        [1] https://stokes.byu.edu/iso/findsym.php
        """
        def identify_wyckpos(sgrp, atoms):
            """
            try to determine suitable Wyckoff positions

            Parameters
            ----------:
            sgrp : str
                space group identifier (including potential suffix)
            atoms : list
                list of atoms in the unit cell. should have equivalent entries
                as SGLattice.base()

            Returns
            -------
            success, atomdict
                "success" is a boolean flag to indicate if equivalents for
                every atom position could be found upon success "atomdict" is a
                dictionary which can be used in a SGLattice definition to
                specify the unit cell. If success==False atomdict is None.
            """
            # get all Wyckpos for this spacegroup
            atomdict = {'atoms': [], 'pos': [], 'occ': [], 'b': []}
            success = True

            elements = set(at[0] for at in atoms)
            # check all atomic species seperately
            for el in elements:
                catoms = list(filter(lambda at: at[0] == el, atoms))
                found = numpy.zeros(len(catoms), dtype=bool)
                # see if atomic positions fit to Wyckoff positions
                for k, wyckpos in wp[sgrp].items():
                    num = int(k[:-1])
                    if num > len(catoms)-sum(found):
                        break
                    parint, poslist, reflcond = wyckpos
                    for f, (dummy, xyz, occ, biso) in zip(found, catoms):
                        if f:
                            continue
                        foundwp, pospar = testwp(parint, poslist[0], xyz)
                        if foundwp:
                            # generate parameters of this Wyckoff position
                            pardict = _get_pardict(parint, pospar)

                            # check if all equivalent positions are also
                            # occupied by the same atom
                            nremoved = 0
                            for p in poslist:
                                pos = eval(p, {}, pardict)
                                pos = SymOp.foldback(pos)
                                for n, entry in enumerate(catoms):
                                    if numpy.allclose(entry[1], pos):
                                        # remove atom from search list
                                        nremoved += 1
                                        found[n] = True
                            if nremoved != num:
                                # not all equivalent positions occupied
                                return False, None
                            else:
                                # add Wyckoff position to output
                                if pospar is None:
                                    atomdict['pos'].append(k)
                                else:
                                    atomdict['pos'].append((k, list(pospar)))
                                atomdict['atoms'].append(el)
                                atomdict['occ'].append(occ)
                                atomdict['b'].append(biso)

                    if num > len(catoms)-sum(found):
                        break
                if len(catoms) != sum(found):
                    success = False
                    atomdict = None
                    break

            return success, atomdict

        # determine possible space groups for unit cell parameters
        systems = []
        possible_sg = []

        # determine possible lattice systems
        for sys in ('cubic', 'hexagonal', 'trigonal:R', 'trigonal:H',
                    'tetragonal', 'orthorhombic', 'monoclinic:b',
                    'monoclinic:c', 'triclinic'):
            freepar, ucpar = sgrp_params[sys]
            ucsys = []
            for par in ucpar:
                if isinstance(par, numbers.Number):
                    ucsys.append(par)
                else:
                    ucsys.append(eval(par, {}, self._parameters))

            if numpy.allclose(tuple(self._parameters.values()), ucsys):
                systems.append(sys)

        # determine suitable space group numbers and names for these families
        for sys in systems:
            freepar, ucpar = sgrp_params[sys]
            splitsys = sys.split(':')
            suf = None
            if len(splitsys) > 1:
                suf = ':' + splitsys[1]
            for sgrange, v in sgrp_sym.items():
                if v[0] == splitsys[0]:
                    for nr in reversed(sgrange):
                        sufs = get_possible_sgrp_suf(nr)
                        if suf is None:
                            if isinstance(sufs, list):
                                for s in sufs:
                                    possible_sg.append((str(nr) + s, freepar))
                            else:
                                possible_sg.append((str(nr), freepar))
                        else:
                            if suf in sufs:
                                possible_sg.append((str(nr) + suf, freepar))
                            else:
                                possible_sg.append((str(nr), freepar))

        # test space groups starting with the highest symmetric one
        if config.VERBOSITY >= config.DEBUG:
            print("XU.materials.SGLattice.findsym: possible space groups: ",
                  possible_sg)

        for sgrp, ucpar in possible_sg:
            success, atoms = identify_wyckpos(sgrp, list(self.base()))
            if success:
                break

        return type(self)(sgrp, *[self._parameters[k] for k in ucpar], **atoms)

    def transform(self, mat, origin):
        """
        Transform the unit cell with the matrix and origin shift given in the
        parameters. This function returns a new instance of SGLattice which
        contains the highest possible symmetry description of the transformed
        unit cell. After the transformation (see [1]) the findsym method is
        used to create the new SGLattice instance.

        Parameters
        ----------
        mat : (3, 3) list, or ndarray, optional
            transformation matrix of the unit cell. The matrix definition aims
            to be consistent with what is used on the Bilbao Crystallographic
            Server [1]. This only defines the linear part, while the origin
            shift is given by origin.
        origin :    (3, ) list, or ndarray
            origin shift of the unit cell [1].

        [1] https://www.cryst.ehu.es/cgi-bin/cryst/programs/nph-doc-trmat
        """
        # transform unit cell dimensions
        ait = numpy.transpose(mat @ self._ai.T)
        a, b, c = [math.VecNorm(ait[i, :]) for i in range(3)]
        al, be, ga = [math.VecAngle(ait[(i+1) % 3, :],
                                    ait[(i+2) % 3, :],
                                    deg=True) for i in range(3)]
        param = (a, b, c, al, be, ga)

        # transform atomic positions and also search neighboring cells to find
        # all equivalent positions
        def recurse_cells(fracpos, outset):
            """
            recursive search in neighboring cells.

            Note that if this function would operate on the Wyckoff position
            level it could use the determinant of the transformation matrix
            (unit cell volume change) to derive an additional abort criterion
            """
            pos = invmat @ fracpos + origin
            pos = SymOp.foldback(pos)
            pos = tuple(numpy.round(pos, config.DIGITS))
            if pos in outset:
                return
            else:
                outset.add(pos)
                g = numpy.mgrid[-1:2, -1:2, -1:2].T.reshape(27, 3).tolist()
                g.remove([0, 0, 0])
                for off in g:
                    recurse_cells(numpy.add(fracpos, off), outset)

        allatoms = {'atoms': [], 'pos': [], 'occ': [], 'b': []}
        invmat = numpy.linalg.inv(mat)
        elements = set(at[0] for at in self.base())
        # check all atomic species seperately
        for el in elements:
            catoms = list(filter(lambda at: at[0] == el, self.base()))
            elset = set()
            for at in catoms:
                eqpos = set()
                recurse_cells(at[1], eqpos)
                for pos in eqpos:
                    if (pos, at[2], at[3]) not in elset:
                        elset.add((pos, at[2], at[3]))
                        allatoms['pos'].append(('1a', pos))
                        allatoms['atoms'].append(el)
                        allatoms['occ'].append(at[2])
                        allatoms['b'].append(at[3])

        p1 = type(self)(1, *param, **allatoms)
        if config.VERBOSITY >= config.DEBUG:
            print("XU.materials.SGLattice: transform via P1: ", p1)
        return p1.findsym()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
a set of  routines to convert Seifert ASCII files to HDF5
in fact there exist two posibilities how the data is stored (depending on the
use detector):

 1. as a simple line scan (using the point detector)
 2. as a map using the PSD

In the first case the data ist stored
"""

import itertools
import os.path
import re

import numpy

from .. import config
from .helper import xu_open

# define some regular expressions
nscans_re = re.compile(r"^&NumScans=\d+")
scan_data_re = re.compile(r"^\#Values\d+")
scan_partab_re = re.compile(r"^\#ScanTableParameter")
novalues_re = re.compile(r"^&NoValues=\d+")
scanaxis_re = re.compile(r"^&ScanAxis=.*")

# some constant regular expressions
re_measparams = re.compile(r"#MeasParameter")
re_rsmparams = re.compile(r"#RsmParameter")
re_data = re.compile(r"#Values")
re_keyvalue = re.compile(r"&\S+=\S")
re_invalidkeyword = re.compile(r"^\d+\S")
re_multiblank = re.compile(r"\s+")
re_position = re.compile(r"&Pos=[+-]*\d*\.\d*")
re_start = re.compile(r"&Start=[+-]*\d*\.\d*")
re_end = re.compile(r"&End=[+-]*\d*\.\d*")
re_step = re.compile(r"&Step=[+-]*\d*\.\d*")
re_time = re.compile(r"&Time=\d*.\d*")
re_stepscan = re.compile(r"^&Start")
re_dataline = re.compile(r"^[+-]*\d*\.\d*")
re_absorber = re.compile(r"^&Axis=A6")


def repair_key(key):
    """
    Repair a key string in the sense that the string is changed in a way that
    it can be used as a valid Python identifier. For that purpose all blanks
    within the string will be replaced by _ and leading numbers get an
    preceeding _.
    """

    if re_invalidkeyword.match(key):
        key = "_" + key

    # now replace all blanks
    key = key.replace(" ", "_")

    return key


class SeifertHeader(object):
    """
    helper class to represent a Seifert (NJA) scan file header
    """

    def __init__(self):
        pass

    def __str__(self):
        ostr = ""
        for k in self.__dict__.keys():
            value = self.__getattribute__(k)
            if isinstance(value, float):
                ostr += k + " = %f\n" % value
            else:
                ostr += k + " = %s\n" % value

        return ostr


class SeifertMultiScan(object):
    """
    Class to parse a Seifert (NJA) multiscan file
    """

    def __init__(self, filename, m_scan, m2, path=""):
        """
        Parse data from a multiscan Seifert file.

        Parameters
        ----------
        filename :  str
            name of the NJA file
        m_scan :    str
            name of the scan axis
        m2 :        str
            name of the second moving motor
        path :      str, optional
            path to the datafile
        """
        self.Filename = os.path.join(path, filename)

        self.nscans = 0  # total number of scans
        self.npscan = 0  # number of points per scan
        self.ctime = 0  # counting time
        self.re_m2 = re.compile(r"^&Axis=%s\s+&Task=Drive" % m2)
        self.re_sm = re.compile(r"^&ScanAxis=%s" % m_scan)
        self.scan_motor_name = m_scan
        self.sec_motor_name = m2

        self.m2_pos = []
        self.sm_pos = []
        self.data = []
        self.n_sm_pos = 0

        with xu_open(self.Filename) as self.fid:
            if config.VERBOSITY >= config.INFO_LOW:
                print("XU.io.SeifertScan: parsing file: %s" % self.Filename)
            self.parse()

    def parse(self):
        self.data = []
        m2_tmppos = None
        self.sm_pos = []
        self.m2_pos = []

        # flag to check if all header information was parsed
        header_complete = False

        for line in self.fid:
            lb = line.decode('ascii').strip()

            # the first thing needed is the number of scans in the file (in
            # file header)
            if nscans_re.match(lb):
                t = lb.split("=")[1]
                self.nscans = int(t)

            if self.re_m2.match(lb):
                t = re_position.findall(lb)[0]
                t = t.split("=")[1]
                m2_tmppos = float(t)

            if novalues_re.match(lb):
                t = lb.split("=")[1]
                self.n_sm_pos = int(t)
                header_complete = True

            if header_complete:
                # append motor positions of second motor
                self.m2_pos.append([[m2_tmppos] * self.n_sm_pos])

                # reset header flag
                header_complete = False
                # read data lines (number of lines determined by number of
                # values)
                datalines = itertools.islice(self.fid, self.n_sm_pos)
                t = numpy.loadtxt(datalines)
                self.data.append(t[:, 1])
                self.sm_pos.append(t[:, 0])

        # after reading all the data
        self.m2_pos = numpy.array(self.m2_pos, dtype=numpy.double)
        self.sm_pos = numpy.array(self.sm_pos, dtype=numpy.double)
        self.data = numpy.array(self.data, dtype=numpy.double)

        self.data.shape = (self.nscans, self.n_sm_pos)
        self.m2_pos.shape = (self.nscans, self.n_sm_pos)
        self.sm_pos.shape = (self.nscans, self.n_sm_pos)


class SeifertScan(object):
    """
    Class to parse a single Seifert (NJA) scan file
    """

    def __init__(self, filename, path=""):
        """
        Constructor for a SeifertScan object.

        Parameters
        ----------
        filename :  str
            a string with the name of the file to read
        path :      str, optional
            path to the datafile
        """
        self.Filename = os.path.join(path, filename)

        self.hdr = SeifertHeader()
        self.data = []
        self.axispos = {}

        with xu_open(self.Filename) as self.fid:
            if config.VERBOSITY >= config.INFO_LOW:
                print("XU.io.SeifertScan: parsing file: %s" % self.Filename)
            self.parse()

        if self.hdr.NumScans != 1:
            self.data.shape = (int(self.data.shape[0] / self.hdr.NoValues),
                               int(self.hdr.NoValues), 2)

    def parse(self):
        if config.VERBOSITY >= config.INFO_ALL:
            print("XU.io.SeifertScan.parse: starting the parser")
        self.data = []
        for line in self.fid:
            lb = line.decode('ascii')
            # remove leading and trailing whitespace and newline characeters
            lb = lb.strip()

            # every line is broken into its content
            llist = re_multiblank.split(lb)
            tmplist = []
            axes = ""
            for e in llist:
                # if the entry is a key value pair
                if re_keyvalue.match(e):
                    (key, value) = e.split("=")
                    # remove leading & from the key
                    key = key[1:]
                    # have to manage malformed key names that cannot be used as
                    # Python identifiers (leading numbers or blanks inside the
                    # name)
                    key = repair_key(key)

                    # try to convert the values to float numbers
                    # leave them as strings if this is not possible
                    try:
                        value = float(value)
                    except ValueError:
                        pass

                    if key == "Axis":
                        axes = value
                        if value not in self.axispos:
                            self.axispos[value] = []
                    elif key == "Pos":
                        self.axispos[axes] += [value, ]

                    self.hdr.__setattr__(key, value)
                else:
                    try:
                        tmplist.append(float(e))
                    except ValueError:
                        pass

            if tmplist != []:
                self.data.append(tmplist)

        # in the end we convert the data list to a numeric array
        self.data = numpy.array(self.data, dtype=float)
        for key in self.axispos:
            self.axispos[key] = numpy.array(self.axispos[key])


def getSeifert_map(filetemplate, scannrs=None, path=".", scantype="map",
                   Nchannels=1280):
    """
    parses multiple Seifert ``*.nja`` files and concatenates the results.  for
    parsing the xrayutilities.io.SeifertMultiScan class is used. The function
    can be used for parsing maps measured with the Meteor1D and point detector.

    Parameters
    ----------
    filetemplate :  str
        template string for the file names, can contain a %d which is replaced
        by the scan number or be a list of filenames
    scannrs :       int or list, optional
        scan number(s)
    path :          str, optional
        common path to the filenames
    scantype :      {'map', 'tsk'}, optional
        type of datafile: can be either 'map' (reciprocal space map measured
        with a regular Seifert job (default)) or 'tsk' (MCA spectra measured
        using the TaskInterpreter)
    Nchannels :     int, optional
        number of channels of the MCA (needed for 'tsk' measurements)

    Returns
    -------
    om, tt, psd :   ndarray
        positions and data as flattened numpy arrays

    Examples
    --------
    >>> om, tt, psd = xrayutilities.io.getSeifert_map("samplename_%d.xrdml",
    >>>                                               [1, 2], path="./data")
    """
    # read raw data and convert to reciprocal space
    om = numpy.zeros(0)
    tt = numpy.zeros(0)
    if scantype == "map":
        psd = numpy.zeros(0)
    else:
        psd = numpy.zeros((0, Nchannels))
    # create scan names
    if scannrs is None:
        files = [filetemplate]
    else:
        files = list()
        if not getattr(scannrs, '__iter__', False):
            scannrs = [scannrs]
        for nr in scannrs:
            files.append(filetemplate % nr)

    # parse files
    for f in files:
        if scantype == "map":
            d = SeifertMultiScan(os.path.join(path, f), 'T', 'O')

            om = numpy.concatenate((om, d.m2_pos.flatten()))
            tt = numpy.concatenate((tt, d.sm_pos.flatten()))
            psd = numpy.concatenate((psd, d.data.flatten()))
        else:  # scantype == "tsk":
            d = SeifertScan(os.path.join(path, f))

            om = numpy.concatenate((om, d.axispos['O'].flatten()))
            tt = numpy.concatenate((tt, d.axispos['T'].flatten()))
            psd = numpy.concatenate((psd, d.data[:, :, 1]))

    return om, tt, psd



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import collections.abc
import copy
import numbers

import numpy

from .. import utilities
from ..materials import Crystal, PseudomorphicMaterial
from ..math import CoordinateTransform, Transform


def _multiply(a, b):
    """
    implement multiplication of SMaterial and MaterialList with integer
    """
    if not isinstance(b, int):
        raise TypeError("unsupported operand type(s) for *: "
                        "'%s' and '%s'" % (type(a), type(b)))
    if b < 1:
        raise ValueError("multiplication factor needs to be positive!")
    m = MaterialList('%d * (%s)' % (b, a.name), a)
    for i in range(b-1):
        m.append(copy.deepcopy(a))
    return m


class SMaterial(object):
    """
    Simulation Material. Extends the xrayutilities Materials by properties
    needed for simulations
    """

    def __init__(self, material, **kwargs):
        """
        initialize a simulation material by specifiying its Material and
        optional other properties

        Parameters
        ----------
        material :  Material (Crystal, or Amorphous)
            Material object containing optical/crystal properties of for the
            simulation; a deepcopy is used internally.
        kwargs :    dict
            optional properties of the material needed for the simulation
        """
        self.name = utilities.makeNaturalName(material.name, check=True)
        self.material = copy.deepcopy(material)
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material):
        self._material = material
        if isinstance(material, Crystal):
            self._structural_params = []
            # make lattice parameters attributes
            for param, value in material.lattice.free_parameters.items():
                self._structural_params.append(param)
                setattr(self, param, value)
            # make attributes from atom positions
            for i, wp in enumerate(material.lattice._wbase):
                if wp[1][1] is not None:
                    for j, p in enumerate(wp[1][1]):
                        name = '_'.join(('at%d' % i, wp[0].name,
                                         wp[1][0], str(j), 'pos'))
                        self._structural_params.append(name)
                        setattr(self, name, p)
            # make attributes from atom occupations
            for i, wp in enumerate(material.lattice._wbase):
                name = '_'.join(('at%d' % i, wp[0].name,
                                 wp[1][0], 'occupation'))
                self._structural_params.append(name)
                setattr(self, name, wp[2])
            # make attributes from Debye waller exponents
            for i, wp in enumerate(material.lattice._wbase):
                name = '_'.join(('at%d' % i, wp[0].name, wp[1][0], 'biso'))
                self._structural_params.append(name)
                setattr(self, name, wp[3])

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if hasattr(self, 'material'):
            if isinstance(self.material, Crystal):
                if name in self.material.lattice.free_parameters:
                    setattr(self.material.lattice, name, value)
                if name.startswith('at'):
                    nsplit = name.split('_')
                    idx = int(nsplit[0][2:])
                    wp = self.material.lattice._wbase[idx]
                    # wyckoff position parameter
                    if nsplit[-1] == 'pos':
                        pidx = int(nsplit[-2])
                        wyckpos = (wp[1][0], list(wp[1][1]))
                        wyckpos[1][pidx] = value
                        self.material.lattice._wbase[idx] = (wp[0], wyckpos,
                                                             wp[2], wp[3])
                    # site occupation
                    if nsplit[-1] == 'occupation':
                        self.material.lattice._wbase[idx] = (wp[0], wp[1],
                                                             value, wp[3])
                    # site DW exponent
                    if nsplit[-1] == 'biso':
                        self.material.lattice._wbase[idx] = (wp[0], wp[1],
                                                             wp[2], value)

    def __radd__(self, other):
        return MaterialList('%s + %s' % (other.name, self.name), other, self)

    def __add__(self, other):
        return MaterialList('%s + %s' % (self.name, other.name), self, other)

    def __mul__(self, other):
        return _multiply(self, other)

    __rmul__ = __mul__

    def __repr__(self):
        s = '{cls}-{name} ('.format(name=self.material.name,
                                    cls=self.__class__.__name__)
        for k in self.__dict__:
            if k not in ('name', '_material', '_structural_params'):
                v = getattr(self, k)
                if isinstance(v, numbers.Number):
                    s += '{key}: {value:.5g}, '.format(key=k, value=v)
                else:
                    s += '{key}: {value}, '.format(key=k, value=v)
        return s + ')'


class MaterialList(collections.abc.MutableSequence):
    """
    class representing the basics of a list of materials for simulations within
    xrayutilities. It extends the built in list type.
    """

    def __init__(self, name, *args):
        if not isinstance(name, str):
            raise TypeError("'name' argument must be a string")
        self.name = name
        self.list = list()
        self.namelist = list()
        self.extend(list(args))

    def check(self, v):
        if not isinstance(v, SMaterial):
            raise TypeError('%s can only contain SMaterial as entries!'
                            % self.__class__.__name__)

    def _set_unique_name(self, v):
        if v.name in self.namelist:
            splitname = v.name.split('_')
            if len(splitname) > 1:
                try:
                    num = int(splitname[-1])
                    basename = '_'.join(splitname[:-1])
                except ValueError:
                    num = 1
                    basename = v.name
            else:
                num = 1
                basename = v.name
            name = '{name}_{num:d}'.format(name=basename, num=num)
            while name in self.namelist:
                num += 1
                name = '{name}_{num:d}'.format(name=basename, num=num)
            v.name = name
        return v.name

    def __len__(self): return len(self.list)

    def __getitem__(self, i): return self.list[i]

    def __delitem__(self, i): del self.list[i]

    def __setitem__(self, i, v):
        self.check(v)
        self.namelist[i] = self._set_unique_name(v)
        self.list[i] = v

    def insert(self, i, v):
        if isinstance(v, MaterialList):
            vs = v
        else:
            vs = [v, ]
        for j, val in enumerate(vs):
            self.check(val)
            self.namelist.insert(i+j, self._set_unique_name(val))
            self.list.insert(i+j, val)

    def __radd__(self, other):
        ml = MaterialList('%s + %s' % (other.name, self.name))
        ml.append(other)
        ml.append(self)
        return ml

    def __add__(self, other):
        ml = MaterialList('%s + %s' % (self.name, other.name))
        ml.append(self)
        ml.append(other)
        return ml

    def __mul__(self, other):
        return _multiply(self, other)

    __rmul__ = __mul__

    def __str__(self):
        layer = ',\n  '.join([str(entry) for entry in self.list])
        s = '{name} [\n  {layer}\n]'.format(name=self.name, layer=layer)
        return s

    def __repr__(self):
        return self.name


class Layer(SMaterial):
    """
    Object describing part of a thin film sample. The properties of a layer
    are :

    Attributes
    ----------
    material :  Material (Crystal or Amorhous)
        an xrayutilties material describing optical and crystal properties of
        the thin film
    thickness : float
        film thickness in angstrom
    """

    _valid_init_kwargs = {'roughness': 'root mean square roughness',
                          'density': 'density in kg/m^3',
                          'relaxation': 'degree of relaxation',
                          'lat_correl': 'lateral correlation length'}

    def __init__(self, material, thickness, **kwargs):
        """
        constructor for the material saving its properties

        Parameters
        ----------
        material :  Material (Crystal or Amorhous)
            an xrayutilties material describing optical and crystal properties
            of the thin film
        thickness : float
            film thickness in angstrom
        kwargs :    dict
            optional keyword arguments with further layer properties.
        roughness : float, optional
            root mean square roughness of the top interface in angstrom
        density :    float, optional
            density of the material in kg/m^3; If not specified the density of
            the material will be used.
        relaxation : float, optional
            the degree of relaxation in case of crystalline thin films
        lat_correl : float, optional
            the lateral correlation length for diffuse reflectivity
            calculations
        """
        utilities.check_kwargs(kwargs, self._valid_init_kwargs,
                               self.__class__.__name__)
        kwargs['thickness'] = thickness
        super().__init__(material, **kwargs)

    def __getattr__(self, name):
        """
        return default values for properties if they were not set
        """
        if name == "density":
            return self.material.density
        elif name == "roughness":
            return 0
        elif name == "lat_correl":
            return numpy.inf
        elif name == "relaxation":
            return 1
        else:
            return super().__getattribute__(name)


class LayerStack(MaterialList):
    """
    extends the built in list type to enable building a stack of Layer by
    various methods.
    """

    def check(self, v):
        if not isinstance(v, Layer):
            raise TypeError('LayerStack can only contain Layer as entries!')


class CrystalStack(LayerStack):
    """
    extends the built in list type to enable building a stack of crystalline
    Layers by various methods.
    """

    def check(self, v):
        super().check(v)
        if not isinstance(v.material, Crystal):
            raise TypeError('CrystalStack can only contain crystalline Layers'
                            ' as entries!')


class GradedLayerStack(CrystalStack):
    """
    generates a sequence of layers with a gradient in chemical composition
    """

    def __init__(self, alloy, xfrom, xto, nsteps, thickness, **kwargs):
        """
        constructor for a graded buffer of the material 'alloy' with chemical
        composition from 'xfrom' to 'xto' with 'nsteps' number of sublayers.
        The total thickness of the graded buffer is 'thickness'

        Parameters
        ----------
        alloy :         function
            Alloy function which allows to create a material with chemical
            composition 'x' by alloy(x)
        xfrom, xto :    float
            chemical composition from the bottom to top
        nsteps :        int
            number of sublayers in the graded buffer
        thickness :     float
            total thickness of the graded stack
        """
        nfrom = alloy(xfrom).name
        nto = alloy(xto).name
        super().__init__('(' + nfrom + '-' + nto + ')')
        for x in numpy.linspace(xfrom, xto, nsteps):
            layer = Layer(alloy(x), thickness/nsteps, **kwargs)
            self.append(layer)


class PseudomorphicStack001(CrystalStack):
    """
    generate a sequence of pseudomorphic crystalline Layers. Surface
    orientation is assumed to be 001 and materials must be cubic/tetragonal.
    """
    trans = Transform(numpy.identity(3))

    def make_epitaxial(self, i):
        layer = self.list[i]
        if i == 0:
            return layer
        psub = self.list[i-1].material
        mpseudo = PseudomorphicMaterial(psub, layer.material, layer.relaxation,
                                        trans=self.trans)
        self.list[i].material = mpseudo

    def __delitem__(self, i):
        del self.list[i]
        for j in range(i, len(self)):
            self.make_epitaxial(j)

    def __setitem__(self, i, v):
        self.check(v)
        self.namelist[i] = self._set_unique_name(v)
        self.list[i] = v
        for j in range(i, len(self)):
            self.make_epitaxial(j)

    def insert(self, i, v):
        if isinstance(v, MaterialList):
            vs = v
        else:
            vs = [v, ]
        for j, val in enumerate(vs):
            self.check(val)
            self.namelist.insert(i+j, self._set_unique_name(val))
            self.list.insert(i+j, copy.copy(val))
            for k in range(i+j, len(self)):
                self.make_epitaxial(k)


class PseudomorphicStack111(PseudomorphicStack001):
    """
    generate a sequence of pseudomorphic crystalline Layers. Surface
    orientation is assumed to be 111 and materials must be cubic.
    """
    trans = CoordinateTransform((1, -1, 0), (1, 1, -2), (1, 1, 1))


class Powder(SMaterial):
    """
    Object describing part of a powder sample. The properties of a powder
    are:

    Attributes
    ----------
    material :   Crystal
        an xrayutilties material (Crystal) describing optical and crystal
        properties of the powder
    volume :     float
        powder's volume (in pseudo units, since only the relative volume enters
        the calculation)

    crystallite_size_lor :      float, optional
        Lorentzian crystallite size fwhm (m)
    crystallite_size_gauss :    float, optional
        Gaussian crystallite size fwhm (m)
    strain_lor :                float, optional
        extra peak width proportional to tan(theta)
    strain_gauss :              float, optional
        extra peak width proportional to tan(theta)
    preferred_orientation :     tuple, optional
        HKL of the preferred orientation
    preferred_orientation_factor : float, optional
        March-Dollase preferred orientation factor: > 1 for platy crystallits ,
        < 1 for rod-like crystallites, and = 1 for random orientation of
        crystallites.
    """

    _valid_init_kwargs = {'crystallite_size_lor': 'Lorentzian cryst. size',
                          'crystallite_size_gauss': 'Gaussian cryst. size',
                          'strain_lor': 'microstrain broadening',
                          'strain_gauss': 'microstrain broadening',
                          'preferred_orientation': 'HKL of pref. orientation',
                          'preferred_orientation_factor':
                          'March-Dollase preferred orientation factor'}

    def __init__(self, material, volume, **kwargs):
        """
        constructor for the material saving its properties

        Parameters
        ----------
        material :      Crystal
            an xrayutilties material (Crystal) describing optical and crystal
            properties of the powder
        volume :        float
            powder's volume (in pseudo units, since only the relative volume
            enters the calculation)
        kwargs :        dict
            optional keyword arguments with further powder properties.
        crystallite_size_lor :      float, optional
            Lorentzian crystallite size fwhm (m)
        crystallite_size_gauss :    float, optional
            Gaussian crystallite size fwhm (m)
        strain_lor, strain_gauss :  float, optional
            extra peak width proportional to tan(theta);
            typically interpreted as microstrain broadening
        """
        utilities.check_kwargs(kwargs, self._valid_init_kwargs,
                               self.__class__.__name__)
        kwargs['volume'] = volume
        super().__init__(material, **kwargs)


class PowderList(MaterialList):
    """
    extends the built in list type to enable building a list of Powder
    by various methods.
    """

    def check(self, v):
        if not isinstance(v, Powder):
            raise TypeError('PowderList can only contain Powder as entries!')



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy

from .. import config, utilities


def coplanar_alphai(qx, qz, en='config', nan_Laue=False):
    """
    calculate coplanar incidence angle from knowledge of the qx (inplane) and
    qz (out of plane) coordinates

    Parameters
    ----------
    qx :    array-like
        inplane momentum transfer component
    qz :    array-like
        out of plane momentum transfer component
    en :    float or str, optional
        x-ray energy (eV). By default the value from the config is used.
    nan_Laue : bool, optional
        set output points inside the Laue zone to 'nan'. default: False

    Returns
    -------
    alphai : array-like
        the incidence angle in degree
    """
    if isinstance(en, str) and en == 'config':
        en = utilities.energy(config.ENERGY)
    k = 2 * numpy.pi / utilities.en2lam(en)
    th = numpy.arcsin(numpy.sqrt(qx**2 + qz**2) / (2 * k))
    ai = numpy.arctan2(qx, qz) + th
    if nan_Laue:
        if isinstance(ai, numpy.ndarray):  # remove positions in Laue zone
            ai[qz < numpy.sqrt(2 * qx * k - qx**2)] = numpy.nan
        else:
            if qz < numpy.sqrt(2 * qx * k - qx**2):
                ai = numpy.nan
    return numpy.degrees(ai)


def coplanar_alphaf(qx, qz, en='config', nan_Laue=False):
    """
    calculate coplanar exit angle from knowledge of the qx (inplane) and qz
    (out of plane) coordinates

    Parameters
    ----------
    qx :    array-like
        inplane momentum transfer component
    qz :    array-like
        out of plane momentum transfer component
    en :    float or str, optional
        x-ray energy (eV). By default the value from the config is used.
    nan_Laue : bool, optional
        set output points inside the Laue zone to 'nan'. default: False

    Returns
    -------
    alphaf : array-like
        the exit angle in degree. points in the Laue zone are set to
        'nan'.
    """
    if isinstance(en, str) and en == 'config':
        en = utilities.energy(config.ENERGY)
    k = 2 * numpy.pi / utilities.en2lam(en)
    th = numpy.arcsin(numpy.sqrt(qx**2 + qz**2) / (2 * k))
    af = th - numpy.arctan2(qx, qz)
    if nan_Laue:
        if isinstance(af, numpy.ndarray):  # remove positions in Laue zone
            af[qz < numpy.sqrt(2 * qx * k - qx**2)] = numpy.nan
        else:
            if qz < numpy.sqrt(2 * qx * k - qx**2):
                af = numpy.nan
    return numpy.degrees(af)


def get_qz(qx, alphai, en='config'):
    """
    calculate the qz position from the qx position and the incidence angle for
    a coplanar diffraction geometry

    Parameters
    ----------
    qx :        array-like
        inplane momentum transfer component
    alphai :    array-like
        incidence angle (deg)
    en :        float or str, optional
        x-ray energy (eV). By default the value from the config is used.

    Returns
    -------
    array-like
        the qz position for the given incidence angle. The output is nan in
        case the inplane momentum transfer can not be reached.
    """
    if isinstance(en, str) and en == 'config':
        en = utilities.energy(config.ENERGY)
    k = 2 * numpy.pi / utilities.en2lam(en)
    ai = numpy.radians(alphai)
    qz = numpy.sqrt(k**2 - (qx + k * numpy.cos(ai))**2) + k * numpy.sin(ai)
    return qz



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy

from .. import config
from . import vector


class Transform(object):

    def __init__(self, matrix):
        self.matrix = matrix
        self._imatrix = None

    @property
    def imatrix(self):
        if self._imatrix is None:
            try:
                self._imatrix = numpy.linalg.inv(self.matrix)
            except numpy.linalg.LinAlgError:
                raise Exception("XU.math.Transform: matrix cannot be inverted"
                                " - seems to be singular")
        return self._imatrix

    def inverse(self, args, rank=1):
        """
        performs inverse transformation a vector, matrix or tensor of rank 4

        Parameters
        ----------
        args :      list or array-like
            object to transform, list or numpy array of shape (..., n)
            (..., n, n), (..., n, n, n, n) where n is the size of the
            transformation matrix.
        rank :      int
            rank of the supplied object. allowed values are 1, 2, and 4
        """
        it = Transform(self.imatrix)
        return it(args, rank)

    def __call__(self, args, rank=1):
        """
        transforms a vector, matrix or tensor of rank 4
        (e.g. elasticity tensor)

        Parameters
        ----------
        args :      list or array-like
            object to transform, list or numpy array of shape (..., n)
            (..., n, n), (..., n, n, n, n) where n is the size of the
            transformation matrix.
        rank :      int
            rank of the supplied object. allowed values are 1, 2, and 4
        """

        m = self.matrix
        if rank == 1:  # argument is a vector
            # out_i = m_ij * args_j
            out = numpy.einsum('ij,...j', m, args)
        elif rank == 2:  # argument is a matrix
            # out_ij = m_ik * m_jl * args_kl
            out = numpy.einsum('ik, jl,...kl', m, m, args)
        elif rank == 4:
            # cp_ijkl = m_in * m_jo * m_kp * m_lq * args_nopq
            out = numpy.einsum('in, jo, kp, lq,...nopq', m, m, m, m, args)

        return out

    def __str__(self):
        ostr = "Transformation matrix:\n"
        ostr += str(self.matrix)
        return ostr


class CoordinateTransform(Transform):

    """
    Create a Transformation object which transforms a point into a new
    coordinate frame. The new frame is determined by the three vectors
    v1/norm(v1), v2/norm(v2) and v3/norm(v3), which need to be orthogonal!
    """

    def __init__(self, v1, v2, v3):
        """
        initialization routine for Coordinate transformation

        Parameters
        ----------
        v1, v2, v3 :     list, tuple or array-like
            new base vectors

        Returns
        -------
        Transform
            An instance of a Transform class
        """
        e1 = vector._checkvec(v1)
        e2 = vector._checkvec(v2)
        e3 = vector._checkvec(v3)

        # normalize base vectors
        e1 = e1 / numpy.linalg.norm(e1)
        e2 = e2 / numpy.linalg.norm(e2)
        e3 = e3 / numpy.linalg.norm(e3)

        # check that the vectors are orthogonal
        t1 = numpy.abs(numpy.dot(e1, e2))
        t2 = numpy.abs(numpy.dot(e1, e3))
        t3 = numpy.abs(numpy.dot(e2, e3))
        if not numpy.allclose((t1, t2, t3), 0):
            raise ValueError("given basis vectors need to be orthogonal!")

        if config.VERBOSITY >= config.INFO_ALL:
            print("XU.math.CoordinateTransform: new basis set: \n"
                  " x: (%5.2f %5.2f %5.2f) \n"
                  " y: (%5.2f %5.2f %5.2f) \n"
                  " z: (%5.2f %5.2f %5.2f)"
                  % (e1[0], e1[1], e1[2], e2[0], e2[1],
                     e2[2], e3[0], e3[1], e3[2]))

        # assemble the transformation matrix
        m = numpy.array([e1, e2, e3])

        Transform.__init__(self, m)


class AxisToZ(CoordinateTransform):

    """
    Creates a coordinate transformation to move a certain axis to the z-axis.
    The rotation is done along the great circle.  The x-axis of the new
    coordinate frame is created to be normal to the new and original z-axis.
    The new y-axis is create in order to obtain a right handed coordinate
    system.
    """

    def __init__(self, newzaxis):
        """
        initialize the CoordinateTransformation to move a certain axis to the
        z-axis

        Parameters
        ----------
        newzaxis :  list or array-like
            new z-axis
        """
        newz = vector._checkvec(newzaxis)

        if numpy.isclose(vector.VecAngle([0, 0, 1], newz), 0):
            newx = [1, 0, 0]
            newy = [0, 1, 0]
            newz = [0, 0, 1]
        elif numpy.isclose(vector.VecAngle([0, 0, 1], -newz), 0):
            newx = [-1, 0, 0]
            newy = [0, 1, 0]
            newz = [0, 0, -1]
        else:
            newx = numpy.cross(newz, [0, 0, 1])
            newy = numpy.cross(newz, newx)

        CoordinateTransform.__init__(self, newx, newy, newz)


class AxisToZ_keepXY(CoordinateTransform):

    """
    Creates a coordinate transformation to move a certain axis to the z-axis.
    The rotation is done along the great circle.  The x-axis/y-axis of the new
    coordinate frame is created to be similar to the old x and y directions.
    This variant of AxisToZ assumes that the new Z-axis has its main component
    along the Z-direction
    """

    def __init__(self, newzaxis):
        """
        initialize the CoordinateTransformation to move a certain axis to the
        z-axis

        Parameters
        ----------
        newzaxis :  list or array-like
            new z-axis
        """
        newz = vector._checkvec(newzaxis)

        if numpy.isclose(vector.VecAngle([0, 0, 1], newz), 0):
            newx = [1, 0, 0]
            newy = [0, 1, 0]
            newz = [0, 0, 1]
        elif numpy.isclose(vector.VecAngle([0, 0, 1], -newz), 0):
            newx = [-1, 0, 0]
            newy = [0, 1, 0]
            newz = [0, 0, -1]
        else:
            newx = numpy.cross(newz, [0, 0, 1])
            newy = numpy.cross(newz, newx)
            # rotate newx and newy to be similar to old directions
            ang = numpy.degrees(numpy.arctan2(newz[0], newz[1]))
            newx = rotarb(newx, newz, ang)
            newy = rotarb(newy, newz, ang)

        CoordinateTransform.__init__(self, newx, newy, newz)


def _sincos(alpha, deg):
    if deg:
        a = numpy.radians(alpha)
    else:
        a = alpha
    return numpy.sin(a), numpy.cos(a)


def XRotation(alpha, deg=True):
    """
    Returns a transform that represents a rotation about the x-axis
    by an angle alpha. If deg=True the angle is assumed to be in
    degree, otherwise the function expects radiants.
    """
    sina, cosa = _sincos(alpha, deg)
    m = numpy.array(
        [[1, 0, 0], [0, cosa, -sina], [0, sina, cosa]], dtype=numpy.double)
    return Transform(m)


def YRotation(alpha, deg=True):
    """
    Returns a transform that represents a rotation about the y-axis
    by an angle alpha. If deg=True the angle is assumed to be in
    degree, otherwise the function expects radiants.
    """
    sina, cosa = _sincos(alpha, deg)
    m = numpy.array(
        [[cosa, 0, sina], [0, 1, 0], [-sina, 0, cosa]], dtype=numpy.double)
    return Transform(m)


def ZRotation(alpha, deg=True):
    """
    Returns a transform that represents a rotation about the z-axis
    by an angle alpha. If deg=True the angle is assumed to be in
    degree, otherwise the function expects radiants.
    """
    sina, cosa = _sincos(alpha, deg)
    m = numpy.array(
        [[cosa, -sina, 0], [sina, cosa, 0], [0, 0, 1]], dtype=numpy.double)
    return Transform(m)


# helper scripts for rotations around arbitrary axis
def tensorprod(vec1, vec2):
    """
    function implements an elementwise multiplication of two vectors
    """
    return vec1[:, numpy.newaxis] * numpy.ones((3, 3)) * vec2[numpy.newaxis, :]


def mycross(vec, mat):
    """
    function implements the cross-product of a vector with each column of a
    matrix
    """
    out = numpy.zeros((3, 3))
    for i in range(3):
        out[:, i] = numpy.cross(vec, mat[:, i])
    return out


def ArbRotation(axis, alpha, deg=True):
    """
    Returns a transform that represents a rotation around an arbitrary axis by
    the angle alpha. positive rotation is anti-clockwise when looking from
    positive end of axis vector

    Parameters
    ----------
    axis :  list or array-like
        rotation axis
    alpha : float
        rotation angle in degree (deg=True) or in rad (deg=False)
    deg :   bool
        determines the input format of ang (default: True)

    Returns
    -------
    Transform
    """
    axis = vector._checkvec(axis)
    e = axis / numpy.linalg.norm(axis)
    if deg:
        rad = numpy.radians(alpha)
    else:
        rad = alpha
    get = tensorprod(e, e)
    rot = get + numpy.cos(rad) * (numpy.identity(3) - get) + \
        numpy.sin(rad) * mycross(e, numpy.identity(3))
    return Transform(rot)


def rotarb(vec, axis, ang, deg=True):
    """
    function implements the rotation around an arbitrary axis by an angle ang
    positive rotation is anti-clockwise when looking from positive end of axis
    vector

    Parameters
    ----------
    vec :   list or array-like
        vector to rotate
    axis :  list or array-like
        rotation axis
    ang :   float
        rotation angle in degree (deg=True) or in rad (deg=False)
    deg :   bool
        determines the input format of ang (default: True)

    Returns
    -------
    rotvec :  rotated vector as numpy.array

    Examples
    --------
    >>> rotarb([1, 0, 0],[0, 0, 1], 90)
    array([  6.12323400e-17,   1.00000000e+00,   0.00000000e+00])
    """
    return ArbRotation(axis, ang, deg)(vec)
